# Copyright (c) 2020 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from blazar import exceptions
from blazar.i18n import _


class MaxLeaseDurationException(exceptions.NotAuthorized):
    code = 400
    msg_fmt = _('Lease duration of %(lease_duration)s seconds must be less '
                'than or equal to the maximum lease duration of '
                '%(max_duration)s seconds.')


class ExternalServiceFilterException(exceptions.BlazarException):
    code = 400
    msg_fmt = _('%(message)s')



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Change tenant to project

Revision ID: 2bcfe76b0474
Revises: 0_1
Create Date: 2014-03-19 18:08:50.825586

"""

# revision identifiers, used by Alembic.
revision = '2bcfe76b0474'
down_revision = '10e34bba18e8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    engine = op.get_bind().engine
    if engine.name == 'sqlite':
        # Only for testing purposes with sqlite
        op.execute('CREATE TABLE tmp_leases as SELECT created_at, updated_at, '
                   'id, name, user_id, tenant_id as project_id, start_date, '
                   'end_date, trust_id FROM leases')
        op.execute('DROP TABLE leases')
        op.execute('ALTER TABLE tmp_leases RENAME TO leases')
        return

    op.alter_column('leases', 'tenant_id', new_column_name='project_id',
                    existing_type=sa.String(length=255))


def downgrade():
    engine = op.get_bind().engine
    if engine.name == 'sqlite':
        # Only for testing purposes with sqlite
        op.execute('CREATE TABLE tmp_leases as SELECT created_at, updated_at, '
                   'id, name, user_id, project_id as tenant_id, start_date, '
                   'end_date, trust_id FROM leases')
        op.execute('DROP TABLE leases')
        op.execute('ALTER TABLE tmp_leases RENAME TO leases')
        return

    op.alter_column('leases', 'project_id', new_column_name='tenant_id',
                    existing_type=sa.String(length=255))



# -*- coding: utf-8 -*-
#
# Author: François Rossigneux <francois.rossigneux@inria.fr>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from random import Random
import retrying

from novaclient import exceptions as nova_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils

from blazar.db import api as db_api
from blazar.db import exceptions as db_ex
from blazar.db import utils as db_utils
from blazar.manager import exceptions as manager_ex
from blazar.plugins import base
from blazar.plugins import oshosts as plugin
from blazar import status
from blazar.utils.openstack import nova
from blazar.utils.openstack import placement
from blazar.utils import plugins as plugins_utils
from blazar.utils import trusts

plugin_opts = [
    cfg.StrOpt('blazar_az_prefix',
               default='blazar_',
               help='Prefix for Availability Zones created by Blazar'),
    cfg.StrOpt('before_end',
               default='',
               help='Actions which we will be taken before the end of '
                    'the lease'),
    cfg.BoolOpt('enable_notification_monitor',
                default=False,
                help='Enable notification-based resource monitoring. '
                     'If it is enabled, the blazar-manager monitors states of '
                     'compute hosts by subscribing to notifications of Nova.'),
    cfg.ListOpt('notification_topics',
                default=['notifications', 'versioned_notifications'],
                help='Notification topics to subscribe to.'),
    cfg.BoolOpt('enable_polling_monitor',
                default=False,
                help='Enable polling-based resource monitoring. '
                     'If it is enabled, the blazar-manager monitors states '
                     'of compute hosts by polling the Nova API.'),
    cfg.IntOpt('polling_interval',
               default=60,
               min=1,
               help='Interval (seconds) of polling for health checking.'),
    cfg.IntOpt('healing_interval',
               default=60,
               min=0,
               help='Interval (minutes) of reservation healing. '
                    'If 0 is specified, the interval is infinite and all the '
                    'reservations in the future is healed at one time.'),
    cfg.BoolOpt('randomize_host_selection',
                default=False,
                help='Allocate hosts for reservations randomly.'),
]

CONF = cfg.CONF
CONF.register_opts(plugin_opts, group=plugin.RESOURCE_TYPE)
LOG = logging.getLogger(__name__)

before_end_options = ['', 'snapshot', 'default']

INSTANCE_DELETION_TIMEOUT = 10 * 60 * 1000  # 10 minutes
QUERY_TYPE_ALLOCATION = 'allocation'


class PhysicalHostPlugin(base.BasePlugin, nova.NovaClientWrapper):
    """Plugin for physical host resource."""
    resource_type = plugin.RESOURCE_TYPE
    title = 'Physical Host Plugin'
    description = 'This plugin starts and shutdowns the hosts.'
    freepool_name = CONF.nova.aggregate_freepool_name
    pool = None
    query_options = {
        QUERY_TYPE_ALLOCATION: ['lease_id', 'reservation_id']
    }

    def __init__(self):
        super(PhysicalHostPlugin, self).__init__(
            username=CONF.os_admin_username,
            password=CONF.os_admin_password,
            user_domain_name=CONF.os_admin_user_domain_name,
            project_name=CONF.os_admin_project_name,
            project_domain_name=CONF.os_admin_project_domain_name)
        self.monitor = PhysicalHostMonitorPlugin()
        self.monitor.register_healing_handler(self.heal_reservations)
        self.placement_client = placement.BlazarPlacementClient()

    def reserve_resource(self, reservation_id, values):
        """Create reservation."""
        host_ids = self.allocation_candidates(values)

        if not host_ids:
            raise manager_ex.NotEnoughHostsAvailable()

        pool = nova.ReservationPool()
        pool_name = reservation_id
        az_name = "%s%s" % (CONF[self.resource_type].blazar_az_prefix,
                            pool_name)
        pool_instance = pool.create(name=pool_name, az=az_name)
        host_rsrv_values = {
            'reservation_id': reservation_id,
            'aggregate_id': pool_instance.id,
            'resource_properties': values['resource_properties'],
            'hypervisor_properties': values['hypervisor_properties'],
            'count_range': values['count_range'],
            'status': 'pending',
            'before_end': values['before_end']
        }
        host_reservation = db_api.host_reservation_create(host_rsrv_values)
        for host_id in host_ids:
            db_api.host_allocation_create({'compute_host_id': host_id,
                                          'reservation_id': reservation_id})
        return host_reservation['id']

    def update_reservation(self, reservation_id, values):
        """Update reservation."""
        reservation = db_api.reservation_get(reservation_id)
        lease = db_api.lease_get(reservation['lease_id'])

        if (not [x for x in values.keys() if x in ['min', 'max',
                                                   'hypervisor_properties',
                                                   'resource_properties']]
                and values['start_date'] >= lease['start_date']
                and values['end_date'] <= lease['end_date']):
            # Nothing to update
            return

        dates_before = {'start_date': lease['start_date'],
                        'end_date': lease['end_date']}
        dates_after = {'start_date': values['start_date'],
                       'end_date': values['end_date']}
        host_reservation = db_api.host_reservation_get(
            reservation['resource_id'])
        self._update_allocations(dates_before, dates_after, reservation_id,
                                 reservation['status'], host_reservation,
                                 values)

        updates = {}
        if 'min' in values or 'max' in values:
            count_range = str(values.get(
                'min', host_reservation['count_range'].split('-')[0])
            ) + '-' + str(values.get(
                'max', host_reservation['count_range'].split('-')[1])
            )
            updates['count_range'] = count_range
        if 'hypervisor_properties' in values:
            updates['hypervisor_properties'] = values.get(
                'hypervisor_properties')
        if 'resource_properties' in values:
            updates['resource_properties'] = values.get(
                'resource_properties')
        if updates:
            db_api.host_reservation_update(host_reservation['id'], updates)

    def on_start(self, resource_id):
        """Add the hosts in the pool."""
        host_reservation = db_api.host_reservation_get(resource_id)
        pool = nova.ReservationPool()
        hosts = []
        for allocation in db_api.host_allocation_get_all_by_values(
                reservation_id=host_reservation['reservation_id']):
            host = db_api.host_get(allocation['compute_host_id'])
            hosts.append(host['service_name'])
        pool.add_computehost(host_reservation['aggregate_id'], hosts)

    def before_end(self, resource_id):
        """Take an action before the end of a lease."""
        host_reservation = db_api.host_reservation_get(resource_id)
        action = host_reservation['before_end']
        if action == 'default':
            action = CONF[plugin.RESOURCE_TYPE].before_end
        if action == 'snapshot':
            pool = nova.ReservationPool()
            client = nova.BlazarNovaClient()
            for host in pool.get_computehosts(
                    host_reservation['aggregate_id']):
                for server in client.servers.list(
                        search_opts={"host": host, "all_tenants": 1}):
                    client.servers.create_image(server=server)

    def on_end(self, resource_id):
        """Remove the hosts from the pool."""
        host_reservation = db_api.host_reservation_get(resource_id)
        reservation_id = host_reservation['reservation_id']
        db_api.host_reservation_update(host_reservation['id'],
                                       {'status': 'completed'})
        allocations = db_api.host_allocation_get_all_by_values(
            reservation_id=reservation_id)
        for allocation in allocations:
            db_api.host_allocation_destroy(allocation['id'])
        pool = nova.ReservationPool()
        for host in pool.get_computehosts(host_reservation['aggregate_id']):
            for server in self.nova.servers.list(
                    search_opts={"host": host, "all_tenants": 1}):
                try:
                    self.nova.servers.delete(server=server)
                except nova_exceptions.NotFound:
                    LOG.info('Could not find server %s, may have been deleted '
                             'concurrently.', server)
                except Exception as e:
                    LOG.exception('Failed to delete %s: %s.', server, str(e))

        # We need to check the deletion is complete before removing the host
        # from the aggregate. See change
        # https://review.opendev.org/c/openstack/nova/+/821423 for details.
        if not self._check_server_deletion(pool, host_reservation):
            LOG.error('Timed out while deleting servers on reservation %s',
                      reservation_id)
            raise manager_ex.ServerDeletionTimeout()

        try:
            pool.delete(host_reservation['aggregate_id'])
        except manager_ex.AggregateNotFound:
            pass

    @retrying.retry(stop_max_delay=INSTANCE_DELETION_TIMEOUT,
                    wait_fixed=5000,  # 5 seconds interval
                    retry_on_result=lambda x: x is False)
    def _check_server_deletion(self, pool, host_reservation):
        servers = []
        for host in pool.get_computehosts(host_reservation['aggregate_id']):
            servers.extend(
                self.nova.servers.list(search_opts={"host": host,
                                                    "all_tenants": 1},
                                       detailed=False))
        if servers:
            LOG.info('Waiting to delete servers: %s ', servers)
            return False
        return True

    def heal_reservations(self, failed_resources, interval_begin,
                          interval_end):
        """Heal reservations which suffer from resource failures.

        :param failed_resources: a list of failed hosts.
        :param interval_begin: start date of the period to heal.
        :param interval_end: end date of the period to heal.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        reservation_flags = {}

        host_ids = [h['id'] for h in failed_resources]
        reservations = db_utils.get_reservations_by_host_ids(host_ids,
                                                             interval_begin,
                                                             interval_end)

        for reservation in reservations:
            if reservation['resource_type'] != plugin.RESOURCE_TYPE:
                continue

            for allocation in [alloc for alloc
                               in reservation['computehost_allocations']
                               if alloc['compute_host_id'] in host_ids]:
                if self._reallocate(allocation):
                    if reservation['status'] == status.reservation.ACTIVE:
                        if reservation['id'] not in reservation_flags:
                            reservation_flags[reservation['id']] = {}
                        reservation_flags[reservation['id']].update(
                            {'resources_changed': True})
                else:
                    if reservation['id'] not in reservation_flags:
                        reservation_flags[reservation['id']] = {}
                    reservation_flags[reservation['id']].update(
                        {'missing_resources': True})

        return reservation_flags

    def _reallocate(self, allocation):
        """Allocate an alternative host.

        :param allocation: allocation to change.
        :return: True if an alternative host was successfully allocated.
        """
        reservation = db_api.reservation_get(allocation['reservation_id'])
        h_reservation = db_api.host_reservation_get(
            reservation['resource_id'])
        lease = db_api.lease_get(reservation['lease_id'])
        pool = nova.ReservationPool()

        # Remove the old host from the aggregate.
        if reservation['status'] == status.reservation.ACTIVE:
            host = db_api.host_get(allocation['compute_host_id'])
            pool.remove_computehost(h_reservation['aggregate_id'],
                                    host['service_name'])

        # Allocate an alternative host.
        start_date = max(datetime.datetime.utcnow(), lease['start_date'])
        new_hostids = self._matching_hosts(
            reservation['hypervisor_properties'],
            reservation['resource_properties'],
            '1-1', start_date, lease['end_date']
        )
        if not new_hostids:
            db_api.host_allocation_destroy(allocation['id'])
            LOG.warning('Could not find alternative host for reservation %s '
                        '(lease: %s).', reservation['id'], lease['name'])
            return False
        else:
            new_hostid = new_hostids.pop()
            db_api.host_allocation_update(allocation['id'],
                                          {'compute_host_id': new_hostid})
            LOG.warning('Resource changed for reservation %s (lease: %s).',
                        reservation['id'], lease['name'])
            if reservation['status'] == status.reservation.ACTIVE:
                # Add the alternative host into the aggregate.
                new_host = db_api.host_get(new_hostid)
                pool.add_computehost(h_reservation['aggregate_id'],
                                     new_host['service_name'])

            return True

    def _get_extra_capabilities(self, host_id):
        extra_capabilities = {}
        raw_extra_capabilities = (
            db_api.host_extra_capability_get_all_per_host(host_id))
        for capability, property_name in raw_extra_capabilities:
            key = property_name
            extra_capabilities[key] = capability.capability_value
        return extra_capabilities

    def get(self, host_id):
        return self.get_computehost(host_id)

    def get_computehost(self, host_id):
        host = db_api.host_get(host_id)
        extra_capabilities = self._get_extra_capabilities(host_id)
        if host is not None and extra_capabilities:
            res = host.copy()
            res.update(extra_capabilities)
            return res
        else:
            return host

    def list_computehosts(self, query=None):
        raw_host_list = db_api.host_list()
        host_list = []
        for host in raw_host_list:
            host_list.append(self.get_computehost(host['id']))
        return host_list

    def create_computehost(self, host_values):
        # TODO(sbauza):
        #  - Exception handling for HostNotFound
        host_id = host_values.pop('id', None)
        host_name = host_values.pop('name', None)
        try:
            trust_id = host_values.pop('trust_id')
        except KeyError:
            raise manager_ex.MissingTrustId()

        host_ref = host_id or host_name
        if host_ref is None:
            raise manager_ex.InvalidHost(host=host_values)

        with trusts.create_ctx_from_trust(trust_id):
            inventory = nova.NovaInventory()
            servers = inventory.get_servers_per_host(host_ref)
            if servers:
                raise manager_ex.HostHavingServers(host=host_ref,
                                                   servers=servers)
            host_details = inventory.get_host_details(host_ref)
            # NOTE(sbauza): Only last duplicate name for same extra capability
            # will be stored
            to_store = set(host_values.keys()) - set(host_details.keys())
            extra_capabilities_keys = to_store
            extra_capabilities = dict(
                (key, host_values[key]) for key in extra_capabilities_keys
            )

            if any([len(key) > 64 for key in extra_capabilities_keys]):
                raise manager_ex.ExtraCapabilityTooLong()

            self.placement_client.create_reservation_provider(
                host_details['hypervisor_hostname'])

            pool = nova.ReservationPool()
            pool.add_computehost(self.freepool_name,
                                 host_details['service_name'])

            host = None
            cantaddextracapability = []
            try:
                if trust_id:
                    host_details.update({'trust_id': trust_id})
                host = db_api.host_create(host_details)
            except db_ex.BlazarDBException as e:
                # We need to rollback
                # TODO(sbauza): Investigate use of Taskflow for atomic
                # transactions
                pool.remove_computehost(self.freepool_name,
                                        host_details['service_name'])
                self.placement_client.delete_reservation_provider(
                    host_details['hypervisor_hostname'])
                raise e
            for key in extra_capabilities:
                values = {'computehost_id': host['id'],
                          'property_name': key,
                          'capability_value': extra_capabilities[key],
                          }
                try:
                    db_api.host_extra_capability_create(values)
                except db_ex.BlazarDBException:
                    cantaddextracapability.append(key)
            if cantaddextracapability:
                raise manager_ex.CantAddExtraCapability(
                    keys=cantaddextracapability,
                    host=host['id'])

            # Check for placement details
            hostname = host_details['hypervisor_hostname']
            rp = self.placement_client.get_resource_provider(hostname)
            if rp is None:
                raise manager_ex.ResourceProviderNotFound(host=hostname)

            inventories = self.placement_client.get_inventory(rp['uuid'])
            for rc, inventory in inventories['inventories'].items():
                resource_inventory = {
                    'computehost_id': host['id'],
                    'resource_class': rc,
                    'total': inventory['total'],
                    'reserved': inventory['reserved'],
                    'min_unit': inventory['min_unit'],
                    'max_unit': inventory['max_unit'],
                    'step_size': inventory['step_size'],
                    'allocation_ratio': inventory['allocation_ratio'],
                }
                db_api.host_resource_inventory_create(resource_inventory)

            traits = self.placement_client.get_traits(rp['uuid'])
            for trait in traits:
                db_api.host_trait_create({
                    'computehost_id': host['id'],
                    'trait': trait,
                })

            return self.get_computehost(host['id'])

    def is_updatable_extra_capability(self, capability, property_name):
        reservations = db_utils.get_reservations_by_host_id(
            capability['computehost_id'], datetime.datetime.utcnow(),
            datetime.date.max)

        for r in reservations:
            plugin_reservation = db_utils.get_plugin_reservation(
                r['resource_type'], r['resource_id'])

            requirements_queries = plugins_utils.convert_requirements(
                plugin_reservation['resource_properties'])

            # TODO(masahito): If all the reservations using the
            # extra_capability can be re-allocated it's okay to update
            # the extra_capability.
            for requirement in requirements_queries:
                # A requirement is of the form "key op value" as string
                if requirement.split(" ")[0] == property_name:
                    return False
        return True

    def update_computehost(self, host_id, values):
        # nothing to update
        if not values:
            return self.get_computehost(host_id)

        cant_update_extra_capability = []
        previous_capabilities = self._get_extra_capabilities(host_id)
        updated_keys = set(values.keys()) & set(previous_capabilities.keys())
        new_keys = set(values.keys()) - set(previous_capabilities.keys())

        for key in updated_keys:
            raw_capability, property_name = next(iter(
                db_api.host_extra_capability_get_all_per_name(host_id, key)))
            capability = {'capability_value': values[key]}

            if self.is_updatable_extra_capability(
                    raw_capability, property_name):
                try:
                    db_api.host_extra_capability_update(
                        raw_capability['id'], capability)
                except (db_ex.BlazarDBException, RuntimeError):
                    cant_update_extra_capability.append(property_name)
            else:
                LOG.info("Capability %s can't be updated because "
                         "existing reservations require it.",
                         property_name)
                cant_update_extra_capability.append(property_name)

        for key in new_keys:
            new_capability = {
                'computehost_id': host_id,
                'property_name': key,
                'capability_value': values[key],
            }
            try:
                db_api.host_extra_capability_create(new_capability)
            except (db_ex.BlazarDBException, RuntimeError):
                cant_update_extra_capability.append(key)

        if cant_update_extra_capability:
            raise manager_ex.CantAddExtraCapability(
                host=host_id, keys=cant_update_extra_capability)

        LOG.info('Extra capabilities on compute host %s updated with %s',
                 host_id, values)
        return self.get_computehost(host_id)

    def delete_computehost(self, host_id):
        host = db_api.host_get(host_id)
        if not host:
            raise manager_ex.HostNotFound(host=host_id)

        with trusts.create_ctx_from_trust(host['trust_id']):
            if db_api.host_allocation_get_all_by_values(
                    compute_host_id=host_id):
                raise manager_ex.CantDeleteHost(
                    host=host_id,
                    msg='The host is reserved.'
                )

            inventory = nova.NovaInventory()
            servers = inventory.get_servers_per_host(
                host['hypervisor_hostname'])
            if servers:
                raise manager_ex.HostHavingServers(
                    host=host['hypervisor_hostname'], servers=servers)

            try:
                pool = nova.ReservationPool()
                pool.remove_computehost(self.freepool_name,
                                        host['service_name'])
                self.placement_client.delete_reservation_provider(
                    host['hypervisor_hostname'])
                # NOTE(sbauza): Extracapabilities will be destroyed thanks to
                #  the DB FK.
                db_api.host_destroy(host_id)
            except db_ex.BlazarDBException as e:
                # Nothing so bad, but we need to alert admins
                # they have to rerun
                raise manager_ex.CantDeleteHost(host=host_id, msg=str(e))

    def list_allocations(self, query):
        hosts_id_list = [h['id'] for h in db_api.host_list()]
        options = self.get_query_options(query, QUERY_TYPE_ALLOCATION)

        hosts_allocations = self.query_allocations(hosts_id_list, **options)
        return [{"resource_id": host, "reservations": allocs}
                for host, allocs in hosts_allocations.items()]

    def get_allocations(self, host_id, query):
        options = self.get_query_options(query, QUERY_TYPE_ALLOCATION)
        host_allocations = self.query_allocations([host_id], **options)
        if host_id not in host_allocations:
            host_allocations = {host_id: []}
        allocs = host_allocations[host_id]
        return {"resource_id": host_id, "reservations": allocs}

    def query_allocations(self, hosts, lease_id=None, reservation_id=None):
        """Return dict of host and its allocations.

        The list element forms
        {
          'host-id': [
                       {
                         'lease_id': lease_id,
                         'id': reservation_id,
                         'start_date': lease_start_date,
                         'end_date': lease_end_date,
                       },
                     ]
        }.
        """
        start = datetime.datetime.utcnow()
        end = datetime.date.max

        # To reduce overhead, this method only executes one query
        # to get the allocation information
        reservations = db_utils.get_reservation_allocations_by_host_ids(
            hosts, start, end, lease_id, reservation_id)
        host_allocs = {h: [] for h in hosts}
        attributes_to_copy = ["id", "lease_id", "start_date", "end_date"]
        for reservation in reservations:
            for host_id in reservation['host_ids']:
                if host_id in host_allocs.keys():
                    host_allocs[host_id].append({
                        k: v for k, v in reservation.items()
                        if k in attributes_to_copy})
        return host_allocs

    def allocation_candidates(self, values):
        self._check_params(values)

        return self._matching_hosts(
            values['hypervisor_properties'],
            values['resource_properties'],
            values['count_range'],
            values['start_date'],
            values['end_date'])

    def _matching_hosts(self, hypervisor_properties, resource_properties,
                        count_range, start_date, end_date):
        """Return the matching hosts (preferably not allocated)

        """
        count_range = count_range.split('-')
        min_host = count_range[0]
        max_host = count_range[1]
        allocated_host_ids = []
        not_allocated_host_ids = []
        filter_array = []
        start_date_with_margin = start_date - datetime.timedelta(
            minutes=CONF.cleaning_time)
        end_date_with_margin = end_date + datetime.timedelta(
            minutes=CONF.cleaning_time)

        # TODO(frossigneux) support "or" operator
        if hypervisor_properties:
            filter_array = plugins_utils.convert_requirements(
                hypervisor_properties)
        if resource_properties:
            filter_array += plugins_utils.convert_requirements(
                resource_properties)
        for host in db_api.reservable_host_get_all_by_queries(filter_array):
            if not db_api.host_allocation_get_all_by_values(
                    compute_host_id=host['id']):
                not_allocated_host_ids.append(host['id'])
            elif db_utils.get_free_periods(
                host['id'],
                start_date_with_margin,
                end_date_with_margin,
                end_date_with_margin - start_date_with_margin
            ) == [
                (start_date_with_margin, end_date_with_margin),
            ]:
                allocated_host_ids.append(host['id'])
        if len(not_allocated_host_ids) >= int(min_host):
            if CONF[self.resource_type].randomize_host_selection:
                Random.shuffle(not_allocated_host_ids)
            return not_allocated_host_ids[:int(max_host)]
        all_host_ids = allocated_host_ids + not_allocated_host_ids
        if len(all_host_ids) >= int(min_host):
            if CONF[self.resource_type].randomize_host_selection:
                Random.shuffle(all_host_ids)
            return all_host_ids[:int(max_host)]
        else:
            return []

    def _convert_int_param(self, param, name):
        """Checks that the parameter is present and can be converted to int."""
        if param is None:
            raise manager_ex.MissingParameter(param=name)
        if strutils.is_int_like(param):
            param = int(param)
        else:
            raise manager_ex.MalformedParameter(param=name)
        return param

    def _check_params(self, values):
        self._validate_min_max_range(values, values.get('min'),
                                     values.get('max'))

        if 'hypervisor_properties' not in values:
            raise manager_ex.MissingParameter(param='hypervisor_properties')
        if 'resource_properties' not in values:
            raise manager_ex.MissingParameter(param='resource_properties')

        if 'before_end' not in values:
            values['before_end'] = 'default'
        if values['before_end'] not in before_end_options:
            raise manager_ex.MalformedParameter(param='before_end')

    def _validate_min_max_range(self, values, min_hosts, max_hosts):
        self._convert_int_param(min_hosts, 'min')
        self._convert_int_param(max_hosts, 'max')
        if min_hosts <= 0 or max_hosts <= 0:
            raise manager_ex.MalformedParameter(
                param='min and max (must be greater than or equal to 1)')
        if max_hosts < min_hosts:
            raise manager_ex.InvalidRange()
        values['count_range'] = str(min_hosts) + '-' + str(max_hosts)

    def _update_allocations(self, dates_before, dates_after, reservation_id,
                            reservation_status, host_reservation, values):
        min_hosts = values.get('min', int(
            host_reservation['count_range'].split('-')[0]))
        max_hosts = values.get(
            'max', int(host_reservation['count_range'].split('-')[1]))
        self._validate_min_max_range(values, min_hosts, max_hosts)
        hypervisor_properties = values.get(
            'hypervisor_properties',
            host_reservation['hypervisor_properties'])
        resource_properties = values.get(
            'resource_properties',
            host_reservation['resource_properties'])
        allocs = db_api.host_allocation_get_all_by_values(
            reservation_id=reservation_id)

        allocs_to_remove = self._allocations_to_remove(
            dates_before, dates_after, max_hosts, hypervisor_properties,
            resource_properties, allocs)

        if (allocs_to_remove and
                reservation_status == status.reservation.ACTIVE):
            raise manager_ex.NotEnoughHostsAvailable()

        kept_hosts = len(allocs) - len(allocs_to_remove)
        if kept_hosts < max_hosts:
            min_hosts = min_hosts - kept_hosts \
                if (min_hosts - kept_hosts) > 0 else 0
            max_hosts = max_hosts - kept_hosts
            host_ids = self._matching_hosts(
                hypervisor_properties, resource_properties,
                str(min_hosts) + '-' + str(max_hosts),
                dates_after['start_date'], dates_after['end_date'])
            if len(host_ids) >= min_hosts:
                new_hosts = []
                pool = nova.ReservationPool()
                for host_id in host_ids:
                    db_api.host_allocation_create(
                        {'compute_host_id': host_id,
                         'reservation_id': reservation_id})
                    new_host = db_api.host_get(host_id)
                    new_hosts.append(new_host['service_name'])
                if reservation_status == status.reservation.ACTIVE:
                    # Add new hosts into the aggregate.
                    pool.add_computehost(host_reservation['aggregate_id'],
                                         new_hosts)
            else:
                raise manager_ex.NotEnoughHostsAvailable()

        for allocation in allocs_to_remove:
            db_api.host_allocation_destroy(allocation['id'])

    def _allocations_to_remove(self, dates_before, dates_after, max_hosts,
                               hypervisor_properties, resource_properties,
                               allocs):
        allocs_to_remove = []
        requested_host_ids = [host['id'] for host in
                              self._filter_hosts_by_properties(
                                  hypervisor_properties, resource_properties)]

        for alloc in allocs:
            if alloc['compute_host_id'] not in requested_host_ids:
                allocs_to_remove.append(alloc)
                continue
            if (dates_before['start_date'] > dates_after['start_date'] or
                    dates_before['end_date'] < dates_after['end_date']):
                reserved_periods = db_utils.get_reserved_periods(
                    alloc['compute_host_id'],
                    dates_after['start_date'],
                    dates_after['end_date'],
                    datetime.timedelta(seconds=1))

                max_start = max(dates_before['start_date'],
                                dates_after['start_date'])
                min_end = min(dates_before['end_date'],
                              dates_after['end_date'])

                if not (len(reserved_periods) == 0 or
                        (len(reserved_periods) == 1 and
                         reserved_periods[0][0] == max_start and
                         reserved_periods[0][1] == min_end)):
                    allocs_to_remove.append(alloc)
                    continue

        kept_hosts = len(allocs) - len(allocs_to_remove)
        if kept_hosts > max_hosts:
            allocs_to_remove.extend(
                [allocation for allocation in allocs
                 if allocation not in allocs_to_remove
                 ][:(kept_hosts - max_hosts)]
            )

        return allocs_to_remove

    def _filter_hosts_by_properties(self, hypervisor_properties,
                                    resource_properties):
        filter = []
        if hypervisor_properties:
            filter += plugins_utils.convert_requirements(hypervisor_properties)
        if resource_properties:
            filter += plugins_utils.convert_requirements(resource_properties)
        if filter:
            return db_api.host_get_all_by_queries(filter)
        else:
            return db_api.host_list()


class PhysicalHostMonitorPlugin(base.BaseMonitorPlugin,
                                nova.NovaClientWrapper):
    """Monitor plugin for physical host resource."""

    # Singleton design pattern
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(PhysicalHostMonitorPlugin, cls).__new__(cls)
            cls._instance.healing_handlers = []
            super(PhysicalHostMonitorPlugin, cls._instance).__init__(
                username=CONF.os_admin_username,
                password=CONF.os_admin_password,
                user_domain_name=CONF.os_admin_user_domain_name,
                project_name=CONF.os_admin_project_name,
                project_domain_name=CONF.os_admin_project_domain_name)
        return cls._instance

    def __init__(self):
        """Do nothing.

        This class uses the Singleton design pattern and an instance of this
        class is generated and initialized in __new__().
        """
        pass

    def register_healing_handler(self, handler):
        self.healing_handlers.append(handler)

    def is_notification_enabled(self):
        """Check if the notification monitor is enabled."""
        return CONF[plugin.RESOURCE_TYPE].enable_notification_monitor

    def get_notification_event_types(self):
        """Get event types of notification messages to handle."""
        return ['service.update']

    def get_notification_topics(self):
        """Get topics of notification to subscribe to."""
        return CONF[plugin.RESOURCE_TYPE].notification_topics

    def notification_callback(self, event_type, payload):
        """Handle a notification message.

        It is used as a callback of a notification-based resource monitor.

        :param event_type: an event type of a notification.
        :param payload: a payload of a notification.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        LOG.trace('Handling a notification...')
        reservation_flags = {}

        data = payload.get('nova_object.data', None)
        if data:
            if data['disabled'] or data['forced_down']:
                failed_hosts = db_api.reservable_host_get_all_by_queries(
                    ['hypervisor_hostname == ' + data['host']])
                if failed_hosts:
                    LOG.warning('%s failed.',
                                failed_hosts[0]['hypervisor_hostname'])
                    reservation_flags = self._handle_failures(failed_hosts)
            else:
                recovered_hosts = db_api.host_get_all_by_queries(
                    ['reservable == 0',
                     'hypervisor_hostname == ' + data['host']])
                if recovered_hosts:
                    db_api.host_update(recovered_hosts[0]['id'],
                                       {'reservable': True})
                    LOG.warning('%s recovered.',
                                recovered_hosts[0]['hypervisor_hostname'])

        return reservation_flags

    def is_polling_enabled(self):
        """Check if the polling monitor is enabled."""
        return CONF[plugin.RESOURCE_TYPE].enable_polling_monitor

    def get_polling_interval(self):
        """Get interval of polling."""
        return CONF[plugin.RESOURCE_TYPE].polling_interval

    def poll(self):
        """Detect and handle resource failures.

        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                 {'missing_resources': True}}
        """
        LOG.trace('Poll...')
        reservation_flags = {}

        failed_hosts, recovered_hosts = self._poll_resource_failures()
        if failed_hosts:
            for host in failed_hosts:
                LOG.warning('%s failed.', host['hypervisor_hostname'])
            reservation_flags = self._handle_failures(failed_hosts)
        if recovered_hosts:
            for host in recovered_hosts:
                db_api.host_update(host['id'], {'reservable': True})
                LOG.warning('%s recovered.', host['hypervisor_hostname'])

        return reservation_flags

    def _poll_resource_failures(self):
        """Check health of hosts by calling Nova Hypervisors API.

        :return: a list of failed hosts, a list of recovered hosts.
        """
        hosts = db_api.host_get_all_by_filters({})
        reservable_hosts = [h for h in hosts if h['reservable'] is True]
        unreservable_hosts = [h for h in hosts if h['reservable'] is False]

        try:
            hvs = self.nova.hypervisors.list()

            failed_hv_ids = [str(hv.id) for hv in hvs
                             if hv.state == 'down' or hv.status == 'disabled']
            failed_hosts = [host for host in reservable_hosts
                            if host['id'] in failed_hv_ids]

            active_hv_ids = [str(hv.id) for hv in hvs
                             if hv.state == 'up' and hv.status == 'enabled']
            recovered_hosts = [host for host in unreservable_hosts
                               if host['id'] in active_hv_ids]
        except Exception as e:
            LOG.exception('Skipping health check. %s', str(e))

        return failed_hosts, recovered_hosts

    def _handle_failures(self, failed_hosts):
        """Handle resource failures.

        :param failed_hosts: a list of failed hosts.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                 {'missing_resources': True}}
        """

        # Update the computehosts table
        for host in failed_hosts:
            try:
                db_api.host_update(host['id'], {'reservable': False})
            except Exception as e:
                LOG.exception('Failed to update %s. %s',
                              host['hypervisor_hostname'], str(e))

        # Heal related reservations
        return self.heal()

    def get_healing_interval(self):
        """Get interval of reservation healing in minutes."""
        return CONF[plugin.RESOURCE_TYPE].healing_interval

    def heal(self):
        """Heal suffering reservations in the next healing interval.

        :return: a dictionary of {reservation id: flags to update}
        """
        reservation_flags = {}
        hosts = db_api.unreservable_host_get_all_by_queries([])

        interval_begin = datetime.datetime.utcnow()
        interval = self.get_healing_interval()
        if interval == 0:
            interval_end = datetime.date.max
        else:
            interval_end = interval_begin + datetime.timedelta(
                minutes=interval)

        for handler in self.healing_handlers:
            reservation_flags.update(handler(hosts,
                                             interval_begin,
                                             interval_end))

        return reservation_flags



# Copyright (c) 2020 University of Chicago.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class BaseFilter(metaclass=abc.ABCMeta):

    enforcement_opts = []

    def __init__(self, conf=None):
        self.conf = conf
        self.register_opts(conf)

    @classmethod
    def register_opts(cls, conf):
        for opt in cls.enforcement_opts:
            conf.register_opt(opt, 'enforcement')

    def __getattr__(self, name):
        func = getattr(self.conf.enforcement, name)
        return func

    @abc.abstractmethod
    def check_create(self, context, lease_values):
        pass

    @abc.abstractmethod
    def check_update(self, context, current_lease_values, new_lease_values):
        pass

    @abc.abstractmethod
    def on_end(self, context, lease_values):
        pass



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
eventlet.monkey_patch(
    os=True, select=True, socket=True, thread=True, time=True)

import flask
from keystonemiddleware import auth_token
from oslo_config import cfg
from oslo_log import log as logging
from oslo_middleware import debug
from stevedore import enabled
from werkzeug import exceptions as werkzeug_exceptions

from blazar.api.v1 import api_version_request
from blazar.api.v1 import request_id
from blazar.api.v1 import request_log
from blazar.api.v1 import utils as api_utils


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

CONF.import_opt('log_exchange', 'blazar.config')


def make_json_error(ex):
    if isinstance(ex, werkzeug_exceptions.HTTPException):
        status_code = ex.code
        description = ex.description
    else:
        status_code = 500
        description = str(ex)
    return api_utils.render({'error': status_code,
                             'error_message': description},
                            status=status_code)


def version_list():
    return api_utils.render({
        "versions": [
            {"id": "v1.0",
             "status": "CURRENT",
             "links": [{"href": "{0}v1".format(flask.request.host_url),
                        "rel": "self"}],
             "min_version": api_version_request.MIN_API_VERSION,
             "max_version": api_version_request.MAX_API_VERSION,
             },
        ],
    }, status="300 Multiple Choices")


def make_app():
    """App builder (wsgi).

    Entry point for Blazar REST API server.
    """
    app = flask.Flask('blazar.api')

    app.route('/', methods=['GET'])(version_list)
    app.route('/versions', methods=['GET'])(version_list)

    LOG.debug("List of plugins: %s", cfg.CONF.manager.plugins)

    plugins = cfg.CONF.manager.plugins + ['leases']
    extension_manager = enabled.EnabledExtensionManager(
        check_func=lambda ext: ext.name in plugins,
        namespace='blazar.api.v1.extensions',
        invoke_on_load=False
        )

    for ext in extension_manager.extensions:
        bp = ext.plugin()
        app.register_blueprint(bp, url_prefix=bp.url_prefix)

    for code in werkzeug_exceptions.default_exceptions:
        app.register_error_handler(code, make_json_error)

    if cfg.CONF.debug and not cfg.CONF.log_exchange:
        LOG.debug('Logging of request/response exchange could be enabled '
                  'using flag --log_exchange')

    if cfg.CONF.log_exchange:
        app.wsgi_app = debug.Debug.factory(app.config)(app.wsgi_app)

    app.wsgi_app = request_id.BlazarReqIdMiddleware(app.wsgi_app)
    app.wsgi_app = request_log.RequestLog(app.wsgi_app)
    app.wsgi_app = auth_token.filter_factory(app.config)(app.wsgi_app)

    return app



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import uuid

from oslo_serialization import jsonutils
from wsme import types as wtypes
from wsme import utils as wutils

from blazar import exceptions


class UuidType(wtypes.UserType):
    """A simple UUID type."""

    basetype = wtypes.text
    name = 'uuid'
    # FIXME(sbauza): When used with wsexpose decorator WSME will try
    # to get the name of the type by accessing it's __name__ attribute.
    # Remove this __name__ attribute once it's fixed in WSME.
    # https://bugs.launchpad.net/wsme/+bug/1265590
    __name__ = name

    def __init__(self, without_dashes=False):
        self.without_dashes = without_dashes

    def validate(self, value):
        try:
            valid_uuid = str(uuid.UUID(value))
            if self.without_dashes:
                valid_uuid = valid_uuid.replace('-', '')
            return valid_uuid
        except (TypeError, ValueError, AttributeError):
            error = 'Value should be UUID format'
            raise ValueError(error)


class IntegerType(wtypes.IntegerType):
    """A simple integer type. Can validate a value range.

    :param minimum: Possible minimum value
    :param maximum: Possible maximum value

    Example::

        Price = IntegerType(minimum=1)

    """

    name = 'integer'
    # FIXME(sbauza): When used with wsexpose decorator WSME will try
    # to get the name of the type by accessing it's __name__ attribute.
    # Remove this __name__ attribute once it's fixed in WSME.
    # https://bugs.launchpad.net/wsme/+bug/1265590
    __name__ = name


class CPUInfo(wtypes.UserType):
    """A type for matching CPU info from hypervisors."""

    basetype = wtypes.text
    name = 'cpuinfo as JSON formated str'

    @staticmethod
    def validate(value):
        # NOTE(sbauza): cpu_info can be very different from one Nova driver to
        #               another. We need to keep this method as generic as
        #               possible, ie. we accept JSONified dict.
        try:
            cpu_info = jsonutils.loads(value)
        except TypeError:
            raise exceptions.InvalidInput(cls=CPUInfo.name, value=value)
        if not isinstance(cpu_info, dict):
            raise exceptions.InvalidInput(cls=CPUInfo.name, value=value)
        return value


class TextOrInteger(wtypes.UserType):
    """A type for matching either text or integer."""

    basetype = wtypes.text
    name = 'textorinteger'

    @staticmethod
    def validate(value):
        # NOTE(sbauza): We need to accept non-unicoded Python2 strings
        if (isinstance(value, str) or isinstance(value, int)):
            return value
        else:
            raise exceptions.InvalidInput(cls=TextOrInteger.name, value=value)


class Datetime(wtypes.UserType):
    """A type for matching unicoded datetime."""

    basetype = wtypes.text
    name = 'datetime'

    # Format must be ISO8601 as default
    format = '%Y-%m-%dT%H:%M:%S.%f'

    def __init__(self, format=None):
        if format:
            self.format = format

    def validate(self, value):
        try:
            datetime.datetime.strptime(value, self.format)
        except ValueError:
            # FIXME(sbauza): Start_date and end_date are given using a specific
            #                format but are shown in default ISO8601, we must
            #                fail back to it for verification
            try:
                wutils.parse_isodatetime(value)
            except ValueError:
                raise exceptions.InvalidInput(cls=Datetime.name, value=value)
        return value



# Copyright (c) 2019 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import neutronclient
from neutronclient.common import exceptions as neutron_exceptions
from oslo_config import cfg
from oslo_config import fixture

from blazar import context
from blazar import tests
from blazar.utils.openstack import exceptions
from blazar.utils.openstack import neutron

CONF = cfg.CONF


class TestBlazarNeutronClient(tests.TestCase):
    def setUp(self):
        super(TestBlazarNeutronClient, self).setUp()
        self.cfg = self.useFixture(fixture.Config(CONF))
        self.context = context
        self.ctx = self.patch(self.context, 'current')

    def test_client_from_kwargs(self):
        kwargs = {
            'auth_url': 'http://foo:8080/identity/v3',
            'region_name': 'RegionTwo',
            'global_request_id': 'req-e19f8f4f-40e7-441e-b776-7b43ed15c7dd'
        }
        client = neutron.BlazarNeutronClient(**kwargs)
        self.assertEqual("http://foo:8080/identity/v3",
                         client.neutron.httpclient.session.auth.auth_url)
        self.assertEqual("RegionTwo", client.neutron.httpclient.region_name)
        self.assertEqual("req-e19f8f4f-40e7-441e-b776-7b43ed15c7dd",
                         client.neutron.httpclient.global_request_id)


class TestFloatingIPPool(tests.TestCase):
    def setUp(self):
        super(TestFloatingIPPool, self).setUp()
        self.mock_net = self.patch(neutronclient.v2_0.client.Client,
                                   'show_network')
        self.mock_net.return_value = {'network': {'id': 'net-id'}}

    def test_init_floatingippool(self):
        client = neutron.FloatingIPPool('net-id')
        self.assertEqual('net-id', client.network_id)
        self.mock_net.assert_called_once_with('net-id')

    def test_init_with_invalid_network_id(self):
        self.mock_net.side_effect = neutron_exceptions.NotFound()

        self.assertRaises(exceptions.FloatingIPNetworkNotFound,
                          neutron.FloatingIPPool, 'invalid-net-id')

    @mock.patch.object(neutronclient.v2_0.client.Client, 'create_floatingip')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'replace_tag')
    def test_create_reserved_floatingip(self, mock_tag, mock_fip):
        mock_fip.return_value = {'floatingip': {'id': 'fip-id'}}

        client = neutron.FloatingIPPool('net-id')
        client.create_reserved_floatingip('subnet-id', '172.24.4.200',
                                          'project-id', 'reservation-id')
        expected_body = {
            'floatingip': {
                'floating_network_id': 'net-id',
                'subnet_id': 'subnet-id',
                'floating_ip_address': '172.24.4.200',
                'project_id': 'project-id'
            }
        }
        mock_fip.assert_called_once_with(expected_body)
        mock_tag.assert_called_once_with(
            'floatingips',
            'fip-id',
            {'tags': ['blazar', 'reservation:reservation-id']}
        )

    @mock.patch.object(neutronclient.v2_0.client.Client, 'list_floatingips')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'update_floatingip')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'delete_floatingip')
    def test_delete_floatingip_with_deleted(self, mock_delete,
                                            mock_update, mock_list):
        mock_list.return_value = {'floatingips': []}

        client = neutron.FloatingIPPool('net-id')
        client.delete_reserved_floatingip('172.24.4.200')

        query = {
            'floating_ip_address': '172.24.4.200',
            'floating_network_id': 'net-id'
        }
        mock_list.assert_called_once_with(**query)
        mock_update.assert_not_called()
        mock_delete.assert_not_called()

    @mock.patch.object(neutronclient.v2_0.client.Client, 'list_floatingips')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'update_floatingip')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'delete_floatingip')
    def test_delete_floatingip_with_associated(self, mock_delete,
                                               mock_update, mock_list):
        mock_list.return_value = {
            'floatingips': [
                {'port_id': 'port-id', 'id': 'fip-id'}
            ]}

        client = neutron.FloatingIPPool('net-id')
        client.delete_reserved_floatingip('172.24.4.200')

        query = {
            'floating_ip_address': '172.24.4.200',
            'floating_network_id': 'net-id'
        }
        mock_list.assert_called_once_with(**query)

        mock_update.assert_called_once_with('fip-id',
                                            {'floatingip': {'port_id': None}})
        mock_delete.assert_called_once_with('fip-id')

    @mock.patch.object(neutronclient.v2_0.client.Client, 'list_floatingips')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'update_floatingip')
    @mock.patch.object(neutronclient.v2_0.client.Client, 'delete_floatingip')
    def test_delete_floatingip_with_deassociated(self, mock_delete,
                                                 mock_update, mock_list):
        mock_list.return_value = {
            'floatingips': [
                {'port_id': None, 'id': 'fip-id'}
            ]}

        client = neutron.FloatingIPPool('net-id')
        client.delete_reserved_floatingip('172.24.4.200')

        query = {
            'floating_ip_address': '172.24.4.200',
            'floating_network_id': 'net-id'
        }
        mock_list.assert_called_once_with(**query)
        mock_update.assert_not_called()
        mock_delete.assert_called_once_with('fip-id')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneclient import client as keystone_client
from keystoneclient import exceptions as keystone_exception
from oslo_config import cfg

from blazar.utils.openstack import base


opts = [
    cfg.StrOpt('identity_service',
               default='identity',
               help='Identity service to use.'),
    cfg.StrOpt('os_region_name',
               default=None,
               help="""
Region name of this node. This is used when picking the URL in the service
catalog.

Possible values:

* Any string representing region name
""")
]

keystone_opts = [
    cfg.StrOpt('endpoint_type',
               default='internal',
               choices=['public', 'admin', 'internal'],
               help='Type of the keystone endpoint to use. This endpoint will '
                    'be looked up in the keystone catalog and should be one '
                    'of public, internal or admin.'),
    cfg.StrOpt('keystone_client_version',
               default='3',
               help='Keystoneclient version'),
]

CONF = cfg.CONF
CONF.register_cli_opts(opts)
CONF.register_opts(keystone_opts)


class BlazarKeystoneClient(object):
    def __init__(self, as_user=False, **kwargs):
        """Return Keystone client for defined in 'identity_service' conf."""
        if as_user:
            client_kwargs = base.client_user_kwargs(**kwargs)
        else:
            client_kwargs = base.client_kwargs(**kwargs)

        client_kwargs.setdefault('version', cfg.CONF.keystone_client_version)
        self.keystone = keystone_client.Client(**client_kwargs)
        self.exceptions = keystone_exception

    def __getattr__(self, name):
        func = getattr(self.keystone, name)
        return func



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps

from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception

from blazar.db import api as db_api
from blazar import exceptions

LOG = logging.getLogger(__name__)


class BaseStatus(object):
    """Base class of status."""

    # All statuses
    ALL = ()

    # Valid status transitions
    NEXT_STATUSES = {}

    @classmethod
    def is_valid_transition(cls, current_status, next_status, **kwargs):
        """Check validity of a status transition.

        :param current_status: Current status
        :param next_status: Next status
        :return: True if the transition is valid
        """

        if next_status not in cls.NEXT_STATUSES[current_status]:
            LOG.warning('Invalid transition from %s to %s.',
                        current_status, next_status)
            return False

        return True


class EventStatus(BaseStatus):
    """Event status class."""

    # Statuses of an event
    UNDONE = 'UNDONE'
    IN_PROGRESS = 'IN_PROGRESS'
    DONE = 'DONE'
    ERROR = 'ERROR'

    ALL = (UNDONE, IN_PROGRESS, DONE, ERROR)

    # Valid status transitions
    NEXT_STATUSES = {
        UNDONE: (IN_PROGRESS,),
        IN_PROGRESS: (DONE, ERROR),
        DONE: (),
        ERROR: ()
    }


class ReservationStatus(BaseStatus):
    """Reservation status class."""

    # Statuses of a reservation
    PENDING = 'pending'
    ACTIVE = 'active'
    DELETED = 'deleted'
    ERROR = 'error'

    ALL = (PENDING, ACTIVE, DELETED, ERROR)

    # Valid status transitions
    NEXT_STATUSES = {
        PENDING: (ACTIVE, DELETED, ERROR),
        ACTIVE: (DELETED, ERROR),
        DELETED: (),
        ERROR: (DELETED,)
    }


class LeaseStatus(BaseStatus):
    """Lease status class."""

    # Stable statuses of a lease
    PENDING = 'PENDING'
    ACTIVE = 'ACTIVE'
    TERMINATED = 'TERMINATED'
    ERROR = 'ERROR'

    STABLE = (PENDING, ACTIVE, TERMINATED, ERROR)

    # Transitional statuses of a lease
    CREATING = 'CREATING'
    STARTING = 'STARTING'
    UPDATING = 'UPDATING'
    TERMINATING = 'TERMINATING'
    DELETING = 'DELETING'

    TRANSITIONAL = (CREATING, STARTING, UPDATING, TERMINATING, DELETING)

    # All statuses
    ALL = STABLE + TRANSITIONAL

    # Valid status transitions
    NEXT_STATUSES = {
        PENDING: (STARTING, UPDATING, DELETING),
        ACTIVE: (TERMINATING, UPDATING, DELETING),
        TERMINATED: (UPDATING, DELETING),
        ERROR: (TERMINATING, UPDATING, DELETING),
        CREATING: (PENDING, DELETING),
        STARTING: (ACTIVE, ERROR, DELETING),
        UPDATING: STABLE + (DELETING,),
        TERMINATING: (TERMINATED, ERROR, DELETING),
        DELETING: (ERROR,)
    }

    @classmethod
    def is_valid_transition(cls, current, next, **kwargs):
        """Check validity of a status transition.

        :param current: Current status
        :param next: Next status
        :return: True if the transition is valid
        """

        if super(LeaseStatus, cls).is_valid_transition(current,
                                                       next, **kwargs):
            if cls.is_valid_combination(kwargs['lease_id'], next):
                return True
            else:
                LOG.warning('Invalid combination of statuses.')

        return False

    @classmethod
    def is_valid_combination(cls, lease_id, status):
        """Validator for the combination of statuses.

        Check if the combination of statuses of lease, reservations and events
        is valid

        :param lease_id: Lease ID
        :param status: Lease status
        :return: True if the combination is valid
        """

        # Validate reservation statuses
        reservations = db_api.reservation_get_all_by_lease_id(lease_id)
        if any([r['status'] not in COMBINATIONS[status]['reservation']
                for r in reservations]):
            return False

        # Validate event statuses
        for event_type in ('start_lease', 'end_lease'):
            event = db_api.event_get_first_sorted_by_filters(
                'lease_id', 'asc',
                {'lease_id': lease_id, 'event_type': event_type}
            )
            if event['status'] not in COMBINATIONS[status][event_type]:
                return False

        return True

    @classmethod
    def is_stable(cls, lease_id):
        """Check if the lease status is stable

        :param lease_id: Lease ID
        :return: True if the status is in (PENDING, ACTIVE, TERMINATED, ERROR)
        """
        lease = db_api.lease_get(lease_id)
        return (lease['status'] in cls.STABLE)

    @classmethod
    def lease_status(cls, transition, result_in, non_fatal_exceptions=[]):
        """Decorator for managing a lease status.

        This checks and updates a lease status before and after executing a
        decorated function.

        :param transition: A status which is set while executing the
                           decorated function.
        :param result_in: A tuple of statuses to which a lease transits after
                          executing the decorated function.
        :param non_fatal_exceptions: A list of exceptions that are non fatal.
                          If one is raised during execution, the lease status
                          will be restored.
        """
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                # Update a lease status
                lease_id = kwargs['lease_id']
                lease = db_api.lease_get(lease_id)
                original_status = lease['status']
                if cls.is_valid_transition(original_status,
                                           transition,
                                           lease_id=lease_id):
                    db_api.lease_update(lease_id,
                                        {'status': transition})
                    LOG.debug('Status of lease %s changed from %s to %s.',
                              lease_id, original_status, transition)
                else:
                    LOG.warning('Aborting %s. '
                                'Invalid lease status transition '
                                'from %s to %s.',
                                func.__name__, original_status,
                                transition)
                    raise exceptions.InvalidStatus

                # Executing the wrapped function
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    with save_and_reraise_exception():
                        if type(e) in non_fatal_exceptions:
                            LOG.exception(
                                'Non-fatal exception during transition '
                                'of lease %s', lease_id)
                            db_api.lease_update(lease_id,
                                                {'status': original_status})
                        else:
                            LOG.exception(
                                'Lease %s went into ERROR status. %s',
                                lease_id, str(e))
                            db_api.lease_update(lease_id,
                                                {'status': cls.ERROR})

                # Update a lease status if it exists
                if db_api.lease_get(lease_id):
                    next_status = cls.derive_stable_status(lease_id)
                    if (next_status in result_in
                            and cls.is_valid_transition(transition,
                                                        next_status,
                                                        lease_id=lease_id)):
                        db_api.lease_update(lease_id,
                                            {'status': next_status})
                        LOG.debug('Status of lease %s changed from %s to %s.',
                                  lease_id, transition, next_status)
                    else:
                        LOG.error('Lease %s went into ERROR status.',
                                  lease_id)
                        db_api.lease_update(lease_id, {'status': cls.ERROR})
                        raise exceptions.InvalidStatus

                return result
            return wrapper
        return decorator

    @classmethod
    def derive_stable_status(cls, lease_id):
        """Derive stable lease status.

        This derives a lease status from statuses of reservations and events.

        :param lease_id: Lease ID
        :return: Derived lease status
        """

        # Possible lease statuses. Key is a tuple of (lease_start event
        # status, lease_end event status)
        possible_statuses = {
            (EventStatus.UNDONE, EventStatus.UNDONE): cls.PENDING,
            (EventStatus.DONE, EventStatus.UNDONE): cls.ACTIVE,
            (EventStatus.DONE, EventStatus.DONE): cls.TERMINATED
        }

        # Derive a lease status from event statuses
        event_statuses = {}
        for event_type in ('start_lease', 'end_lease'):
            event = db_api.event_get_first_sorted_by_filters(
                'lease_id', 'asc',
                {'lease_id': lease_id, 'event_type': event_type}
            )
            event_statuses[event_type] = event['status']
        try:
            status = possible_statuses[(event_statuses['start_lease'],
                                        event_statuses['end_lease'])]
        except KeyError:
            status = cls.ERROR

        # Check the combination of statuses.
        if cls.is_valid_combination(lease_id, status):
            return status
        else:
            return cls.ERROR


COMBINATIONS = {
    LeaseStatus.CREATING: {
        'reservation': (ReservationStatus.PENDING,),
        'start_lease': (EventStatus.UNDONE,),
        'end_lease': (EventStatus.UNDONE,)
    },
    LeaseStatus.PENDING: {
        'reservation': (ReservationStatus.PENDING,),
        'start_lease': (EventStatus.UNDONE,),
        'end_lease': (EventStatus.UNDONE,)
    },
    LeaseStatus.STARTING: {
        'reservation': (ReservationStatus.PENDING,
                        ReservationStatus.ACTIVE,
                        ReservationStatus.ERROR),
        'start_lease': (EventStatus.IN_PROGRESS,),
        'end_lease': (EventStatus.UNDONE,)
    },
    LeaseStatus.ACTIVE: {
        'reservation': (ReservationStatus.ACTIVE,),
        'start_lease': (EventStatus.DONE,),
        'end_lease': (EventStatus.UNDONE,)
    },
    LeaseStatus.TERMINATING: {
        'reservation': (ReservationStatus.ACTIVE,
                        ReservationStatus.DELETED,
                        ReservationStatus.ERROR),
        'start_lease': (EventStatus.DONE,
                        EventStatus.ERROR),
        'end_lease': (EventStatus.IN_PROGRESS,)
    },
    LeaseStatus.TERMINATED: {
        'reservation': (ReservationStatus.DELETED,),
        'start_lease': (EventStatus.DONE,),
        'end_lease': (EventStatus.DONE,)
    },
    LeaseStatus.DELETING: {
        'reservation': ReservationStatus.ALL,
        'start_lease': (EventStatus.UNDONE,
                        EventStatus.DONE,
                        EventStatus.ERROR),
        'end_lease': (EventStatus.UNDONE,
                      EventStatus.DONE,
                      EventStatus.ERROR)
    },
    LeaseStatus.UPDATING: {
        'reservation': ReservationStatus.ALL,
        'start_lease': (EventStatus.UNDONE,
                        EventStatus.DONE,
                        EventStatus.ERROR),
        'end_lease': (EventStatus.UNDONE,
                      EventStatus.DONE,
                      EventStatus.ERROR)
    },
    LeaseStatus.ERROR: {
        'reservation': ReservationStatus.ERROR,
        'start_lease': (EventStatus.DONE,
                        EventStatus.ERROR),
        'end_lease': (EventStatus.UNDONE,
                      EventStatus.ERROR)
    }
}

event = EventStatus
reservation = ReservationStatus
lease = LeaseStatus



# Copyright 2019 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add_floatingip_reservation

Revision ID: f4084140f608
Revises: e069c014356d
Create Date: 2019-02-25 06:25:22.038890

"""

# revision identifiers, used by Alembic.
revision = 'f4084140f608'
down_revision = 'e069c014356d'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('floatingip_allocations',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('floatingip_id',
                              sa.String(length=36), nullable=True),
                    sa.Column('reservation_id',
                              sa.String(length=36), nullable=True),

                    sa.ForeignKeyConstraint(['floatingip_id'],
                                            ['floatingips.id'], ),
                    sa.ForeignKeyConstraint(['reservation_id'],
                                            ['reservations.id'], ),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('floatingip_reservations',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('reservation_id',
                              sa.String(length=36), nullable=True),
                    sa.Column('network_id',
                              sa.String(length=255), nullable=False),
                    sa.Column('amount', sa.Integer(), nullable=False),

                    sa.ForeignKeyConstraint(['reservation_id'],
                                            ['reservations.id'], ),
                    sa.PrimaryKeyConstraint('id'))

    op.create_table('required_floatingips',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('address',
                              sa.String(length=255), nullable=False),
                    sa.Column('floatingip_reservation_id',
                              sa.String(length=36), nullable=True),
                    sa.ForeignKeyConstraint(['floatingip_reservation_id'],
                                            ['floatingip_reservations.id'], ),
                    sa.PrimaryKeyConstraint('id'))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('required_floatingips')
    op.drop_table('floatingip_reservations')
    op.drop_table('floatingip_allocations')
    # ### end Alembic commands ###



# Copyright 2019 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add-floatingip

Revision ID: e069c014356d
Revises: 9593f3656974
Create Date: 2019-01-10 10:15:47.874521

"""

# revision identifiers, used by Alembic.
revision = 'e069c014356d'
down_revision = '9593f3656974'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('floatingips',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('floating_network_id', sa.String(length=255),
                              nullable=False),
                    sa.Column('subnet_id', sa.String(length=255),
                              nullable=False),
                    sa.Column('floating_ip_address', sa.String(length=255),
                              nullable=False),
                    sa.Column('reservable', sa.Boolean(),
                              server_default=sa.text('true'), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('subnet_id', 'floating_ip_address'))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('floatingips')
    # ### end Alembic commands ###



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import wsme
from wsme import types as wtypes

from blazar.api.v2.controllers import types


class _Base(wtypes.DynamicBase):

    # NOTE(sbauza): That does respect ISO8601 but with a different sep (' ')
    created_at = types.Datetime('%Y-%m-%d %H:%M:%S.%f')
    "The time in UTC at which the object is created"

    updated_at = types.Datetime('%Y-%m-%d %H:%M:%S.%f')
    "The time in UTC at which the object is updated"

    def as_dict(self):
        cls = type(self)
        valid_keys = [item for item in dir(cls)
                      if item not in dir(_Base)
                      and wtypes.iswsattr(getattr(cls, item))]

        if 'self' in valid_keys:
            valid_keys.remove('self')
        return self.as_dict_from_keys(valid_keys)

    def as_dict_from_keys(self, keys):
        res = {}
        for key in keys:
            value = getattr(self, key, wsme.Unset)
            if value != wsme.Unset:
                res[key] = value
        return res

    @classmethod
    def convert(cls, rpc_obj):
        obj = cls(**rpc_obj)
        return obj



# -*- encoding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for API /os-hosts/ methods
"""
from oslo_utils import uuidutils

from blazar.tests import api
from blazar.utils import trusts


def fake_computehost(**kw):
    return {
        'id': kw.get('id', '1'),
        'hypervisor_hostname': kw.get('hypervisor_hostname', 'host01'),
        'hypervisor_type': kw.get('hypervisor_type', 'QEMU'),
        'vcpus': kw.get('vcpus', 1),
        'hypervisor_version': kw.get('hypervisor_version', 1000000),
        'trust_id': kw.get('trust_id',
                           '35b17138-b364-4e6a-a131-8f3099c5be68'),
        'memory_mb': kw.get('memory_mb', 8192),
        'local_gb': kw.get('local_gb', 50),
        'cpu_info': kw.get('cpu_info',
                           "{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                           "\"arch\": \"x86_64\", \"features\": [],"
                           " \"topology\": {\"cores\": 1}}",
                           ),
        'extra_capas': kw.get('extra_capas',
                              {'vgpus': 2, 'fruits': 'bananas'})
    }


def fake_computehost_request_body(include=[], **kw):
    computehost_body = fake_computehost(**kw)
    computehost_body['name'] = kw.get('name',
                                      computehost_body['hypervisor_hostname'])
    include.append('name')
    include.append('extra_capas')
    return dict((key, computehost_body[key])
                for key in computehost_body if key in include)


def fake_computehost_from_rpc(**kw):
    # NOTE(sbauza): Extra capabilites are returned as extra key/value pairs
    #               from the Manager when searching from a specific node.
    computehost = fake_computehost(**kw)
    extra_capas = computehost.pop('extra_capas', None)
    if extra_capas is not None:
        computehost.update(extra_capas)
    return computehost


def fake_trust(id=fake_computehost()['trust_id']):
    return type('Trust', (), {
        'id': id,
    })


class TestIncorrectHostFromRPC(api.APITest):

    def setUp(self):
        super(TestIncorrectHostFromRPC, self).setUp()

        self.path = '/os-hosts'
        self.patch(
            self.hosts_rpcapi, 'list_computehosts').return_value = [
                fake_computehost_from_rpc(hypervisor_type=1)
            ]

        self.headers = {'X-Roles': 'admin'}

    def test_bad_list(self):
        expected = {
            'error_code': 400,
            'error_message': "Invalid input",
            'error_name': 400
        }
        response = self.get_json(self.path, expect_errors=True,
                                 headers=self.headers)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])


class TestListHosts(api.APITest):

    def setUp(self):
        super(TestListHosts, self).setUp()

        self.path = '/os-hosts'
        self.patch(self.hosts_rpcapi, 'list_computehosts').return_value = []

        self.headers = {'X-Roles': 'admin'}

    def test_empty(self):
        response = self.get_json(self.path, headers=self.headers)
        self.assertEqual([], response)

    def test_one(self):
        self.patch(
            self.hosts_rpcapi, 'list_computehosts'
        ).return_value = [fake_computehost_from_rpc(id=1)]

        response = self.get_json(self.path, headers=self.headers)
        self.assertEqual([fake_computehost(id=1)], response)

    def test_multiple(self):
        id1 = str('1')
        id2 = str('2')
        self.patch(
            self.hosts_rpcapi, 'list_computehosts').return_value = [
                fake_computehost_from_rpc(id=id1),
                fake_computehost_from_rpc(id=id2)
            ]
        response = self.get_json(self.path, headers=self.headers)
        self.assertEqual([fake_computehost(id=id1), fake_computehost(id=id2)],
                         response)

    def test_rpc_exception_list(self):
        def fake_list_computehosts(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.hosts_rpcapi, 'list_computehosts'
        ).side_effect = fake_list_computehosts
        response = self.get_json(self.path, headers=self.headers,
                                 expect_errors=True)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestShowHost(api.APITest):

    def setUp(self):
        super(TestShowHost, self).setUp()

        self.id1 = str('1')
        self.path = '/os-hosts/{0}'.format(self.id1)
        self.patch(
            self.hosts_rpcapi, 'get_computehost'
        ).return_value = fake_computehost_from_rpc(id=self.id1)

        self.headers = {'X-Roles': 'admin'}

    def test_one(self):
        response = self.get_json(self.path, headers=self.headers)
        self.assertEqual(fake_computehost(id=self.id1), response)

    def test_empty(self):
        expected = {
            'error_code': 404,
            'error_message': "Object with {{'host_id': "
                             "{0}}} not found".format(self.id1),
            'error_name': 404
        }
        self.patch(self.hosts_rpcapi, 'get_computehost').return_value = None
        response = self.get_json(self.path, expect_errors=True,
                                 headers=self.headers)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_rpc_exception_get(self):
        def fake_get_computehost(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': "Nah...",
            'error_name': 500
        }
        self.patch(
            self.hosts_rpcapi, 'get_computehost'
        ).side_effect = fake_get_computehost
        response = self.get_json(self.path, expect_errors=True,
                                 headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestCreateHost(api.APITest):

    def setUp(self):
        super(TestCreateHost, self).setUp()

        self.id1 = str(uuidutils.generate_uuid())
        self.fake_computehost = fake_computehost(id=self.id1)
        self.fake_computehost_body = fake_computehost_request_body(id=self.id1)
        self.path = '/os-hosts'
        self.patch(
            self.hosts_rpcapi, 'create_computehost'
        ).return_value = fake_computehost_from_rpc(id=self.id1)

        self.headers = {'X-Roles': 'admin'}

        self.trusts = trusts
        self.patch(self.trusts, 'create_trust').return_value = fake_trust()

    def test_create_one(self):
        response = self.post_json(self.path, self.fake_computehost_body,
                                  headers=self.headers)
        self.assertEqual(201, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(self.fake_computehost, response.json)

    def test_create_wrong_attr(self):
        expected = {
            "error_name": 400,
            "error_message": "Invalid input for field/attribute name. ",
            "error_code": 400
        }

        response = self.post_json(self.path,
                                  fake_computehost_request_body(name=1),
                                  expect_errors=True, headers=self.headers)
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_create_with_empty_body(self):
        expected = {
            "error_name": 500,
            "error_message": "'NoneType' object has no attribute 'as_dict'",
            "error_code": 500
        }

        response = self.post_json(self.path, None, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_empty_response(self):
        expected = {
            'error_code': 500,
            'error_message': "Host can't be created",
            'error_name': 500
        }
        self.patch(self.hosts_rpcapi, 'create_computehost').return_value = None
        response = self.post_json(self.path, self.fake_computehost_body,
                                  expect_errors=True, headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_rpc_exception_create(self):
        def fake_create_computehost(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': u"Nah...",
            'error_name': 500
        }
        self.patch(
            self.hosts_rpcapi, 'create_computehost'
        ).side_effect = fake_create_computehost
        response = self.post_json(self.path, self.fake_computehost_body,
                                  expect_errors=True, headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestUpdateHost(api.APITest):

    def setUp(self):
        super(TestUpdateHost, self).setUp()

        self.id1 = str('1')
        self.fake_computehost = fake_computehost(id=self.id1, name='updated')
        self.fake_computehost_body = fake_computehost_request_body(
            exclude=['reservations', 'events'],
            id=self.id1,
            name='updated'
        )
        self.path = '/os-hosts/{0}'.format(self.id1)
        self.patch(
            self.hosts_rpcapi, 'update_computehost'
        ).return_value = fake_computehost_from_rpc(id=self.id1, name='updated')

        self.headers = {'X-Roles': 'admin'}

    def test_update_one(self):
        response = self.put_json(self.path, fake_computehost_request_body(
                                 exclude=['trust_id']),
                                 headers=self.headers)
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(self.fake_computehost, response.json)

    def test_update_with_empty_body(self):
        expected = {
            "error_name": 500,
            "error_message": "'NoneType' object has no attribute 'as_dict'",
            "error_code": 500
        }

        response = self.put_json(self.path, None, expect_errors=True,
                                 headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_empty_response(self):
        expected = {
            'error_code': 404,
            'error_message': "Object with {{'host_id': "
                             "{0}}} not found".format(self.id1),
            'error_name': 404
        }
        self.patch(self.hosts_rpcapi, 'update_computehost').return_value = None
        response = self.put_json(self.path, self.fake_computehost_body,
                                 expect_errors=True, headers=self.headers)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)

    def test_rpc_exception_update(self):
        def fake_update_computehost(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': u"Nah...",
            'error_name': 500
        }
        self.patch(
            self.hosts_rpcapi, 'update_computehost'
        ).side_effect = fake_update_computehost
        response = self.put_json(self.path, self.fake_computehost_body,
                                 expect_errors=True, headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)


class TestDeleteHost(api.APITest):

    def setUp(self):
        super(TestDeleteHost, self).setUp()

        self.id1 = str('1')
        self.path = '/os-hosts/{0}'.format(self.id1)
        self.patch(self.hosts_rpcapi, 'delete_computehost')
        self.headers = {'X-Roles': 'admin'}

    def test_delete_one(self):
        response = self.delete(self.path, headers=self.headers)
        self.assertEqual(204, response.status_int)
        self.assertIsNone(response.content_type)
        self.assertEqual(b'', response.body)

    def test_delete_not_existing_computehost(self):
        def fake_delete_computehost(*args, **kwargs):
            raise TypeError("Nah...")
        expected = {
            'error_code': 404,
            'error_message': "not found",
            'error_name': 404
        }
        self.patch(
            self.hosts_rpcapi, 'delete_computehost'
        ).side_effect = fake_delete_computehost
        response = self.delete(self.path, expect_errors=True,
                               headers=self.headers)
        self.assertEqual(404, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected["error_name"], response.json["error_name"])
        self.assertIn(expected["error_message"],
                      response.json["error_message"])
        self.assertEqual(expected["error_code"], response.json["error_code"])

    def test_rpc_exception_delete(self):
        def fake_delete_computehost(*args, **kwargs):
            raise Exception("Nah...")
        expected = {
            'error_code': 500,
            'error_message': u"Nah...",
            'error_name': 500
        }
        self.patch(
            self.hosts_rpcapi, 'delete_computehost'
        ).side_effect = fake_delete_computehost
        response = self.delete(self.path, expect_errors=True,
                               headers=self.headers)
        self.assertEqual(500, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(expected, response.json)



# Copyright (c) 2019 NTT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.manager.floatingips import rpcapi as manager_rpcapi
from blazar import policy
from blazar.utils import trusts


class API(object):
    def __init__(self):
        self.manager_rpcapi = manager_rpcapi.ManagerRPCAPI()

    @policy.authorize('floatingips', 'get')
    def get_floatingips(self):
        """List all existing floatingip."""
        return self.manager_rpcapi.list_floatingips()

    @policy.authorize('floatingips', 'post')
    @trusts.use_trust_auth()
    def create_floatingip(self, data):
        """Create new floatingip.

        :param data: New floatingip characteristics.
        :type data: dict
        """

        return self.manager_rpcapi.create_floatingip(data)

    @policy.authorize('floatingips', 'get')
    def get_floatingip(self, floatingip_id):
        """Get floatingip by its ID.

        :param floatingip_id: ID of the floatingip in Blazar DB.
        :type floatingip_id: str
        """
        return self.manager_rpcapi.get_floatingip(floatingip_id)

    @policy.authorize('floatingips', 'delete')
    def delete_floatingip(self, floatingip_id):
        """Delete specified floatingip.

        :param floatingip_id: ID of the floatingip in Blazar DB.
        :type floatingip_id: str
        """
        self.manager_rpcapi.delete_floatingip(floatingip_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from blazar.api.v1 import utils as api_utils
from blazar import exceptions


def check_exists(get_function, object_id=None, **get_args):
    """Check object exists.

    :param get_function: Method to call to get object.
    :type get_function: function
    :param object_id: ID of the object to get.
    :type object_id: str
    :param get_args: Other params to pass to the get_function method.
    :type get_args: dict
    """
    def decorator(func):
        """Decorate method to check object existing."""
        if object_id is not None:
            get_args['id'] = object_id

        @functools.wraps(func)
        def handler(*args, **kwargs):
            """Decorator handler."""

            get_kwargs = {}
            for k, v in get_args.items():
                get_kwargs[k] = kwargs[v]

            try:
                obj = get_function(**get_kwargs)
            except exceptions.NotFound:
                obj = None
            if obj is None:
                e = exceptions.NotFound(object=get_kwargs)
                return api_utils.not_found(e)

            return func(*args, **kwargs)

        return handler

    return decorator



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Mirantis Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Tests for database migrations. This test case reads the configuration
file test_migrations.conf for database connection settings
to use in the tests. For each connection found in the config file,
the test case runs a series of test cases to ensure that migrations work
properly.

There are also "opportunistic" tests for both mysql and postgresql in here,
which allows testing against mysql and pg in a properly configured unit
test environment.

For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost.
The test will then use that db and u/p combo to run the tests.

For postgres on Ubuntu this can be done with the following commands:

sudo -u postgres psql
postgres=# create user openstack_citest with createdb login password
      'openstack_citest';
postgres=# create database openstack_citest with owner openstack_citest;

"""

from oslo_config import cfg
import sqlalchemy

from blazar.tests.db import migration

CONF = cfg.CONF


class TestMigrations(migration.BaseWalkMigrationTestCase,
                     migration.CommonTestsMixIn):
    """Test alembic migrations."""

    # This variables are used by BaseWalkMigrationTestCase in order to perform
    # the opportunistic testing
    USER = "openstack_citest"
    PASSWD = "openstack_citest"
    DATABASE = "openstack_citest"

    def get_table(self, engine, name):
        """Returns an sqlalchemy table dynamically from db.

        Needed because the models don't work for us in migrations
        as models will be far out of sync with the current data.
        """
        metadata = sqlalchemy.MetaData()
        metadata.bind = engine
        return sqlalchemy.Table(name, metadata, autoload_with=metadata.bind)

    def assertTableExists(self, engine, table):
        metadata = sqlalchemy.MetaData()
        metadata.reflect(bind=engine)
        self.assertIn(table, metadata.tables)

    def assertColumnExists(self, engine, table, column):
        t = self.get_table(engine, table)
        self.assertIn(column, t.c)

    def assertColumnsExists(self, engine, table, columns):
        for column in columns:
            self.assertColumnExists(engine, table, column)

    def assertColumnCount(self, engine, table, columns):
        t = self.get_table(engine, table)
        self.assertEqual(len(t.columns), len(columns))

    def assertColumnNotExists(self, engine, table, column):
        t = self.get_table(engine, table)
        self.assertNotIn(column, t.c)

    def assertIndexExists(self, engine, table, index):
        t = self.get_table(engine, table)
        index_names = [idx.name for idx in t.indexes]
        self.assertIn(index, index_names)

    def assertIndexMembers(self, engine, table, index, members):
        self.assertIndexExists(engine, table, index)

        t = self.get_table(engine, table)
        index_columns = None
        for idx in t.indexes:
            if idx.name == index:
                index_columns = idx.columns.keys()
                break

        self.assertEqual(sorted(members), sorted(index_columns))

    def _check_0_1(self, engine, data):
        self.assertTableExists(engine, 'computehosts')
        self.assertTableExists(engine, 'leases')
        self.assertTableExists(engine, 'reservations')
        self.assertTableExists(engine, 'computehost_extra_capabilities')
        self.assertTableExists(engine, 'events')
        self.assertTableExists(engine, 'computehost_allocations')
        self.assertTableExists(engine, 'computehost_reservations')

    def _check_10e34bba18e8(self, engine, data):
        self.assertColumnExists(engine, 'computehosts', 'service_name')

    def _check_2bcfe76b0474(self, engine, data):
        self.assertColumnExists(engine, 'leases', 'project_id')
        self.assertColumnNotExists(engine, 'leases', 'tenant_id')

    def _pre_upgrade_1fd6c2eded89(self, engine):
        data = [{
            'id': '1',
            'hypervisor_hostname': 'host01',
            'hypervisor_type': 'QEMU',
            'hypervisor_version': 1000000,
            'service_name': 'host01',
            'vcpus': 1,
            'memory_mb': 8192,
            'local_gb': 50,
            'cpu_info': "{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                        "\"arch\": \"x86_64\", \"features\": [],"
                        " \"topology\": {\"cores\": 1}}",
            'extra_capas': {'vgpus': 2}},
            {'id': '2',
             'hypervisor_hostname': 'host01',
             'hypervisor_type': 'QEMU',
             'hypervisor_version': 1000000,
             'service_name': 'host02',
             'vcpus': 1,
             'memory_mb': 8192,
             'local_gb': 50,
             'cpu_info': "{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                         "\"arch\": \"x86_64\", \"features\": [],"
                         " \"topology\": {\"cores\": 1}}",
             'extra_capas': {'vgpus': 2}}]
        computehosts_table = self.get_table(engine, 'computehosts')
        # pylint: disable=E1120
        engine.execute(computehosts_table.insert(), data)
        return data

    def _check_1fd6c2eded89(self, engine, data):
        self.assertColumnExists(engine, 'computehosts', 'trust_id')

        metadata = sqlalchemy.MetaData()
        metadata.bind = engine
        computehosts_table = self.get_table(engine, 'computehosts')

        all_computehosts = computehosts_table.select().execute()
        for computehost in all_computehosts:
            self.assertIn(computehost.trust_id, ['', None])

        data = {'id': '3',
                'hypervisor_hostname': 'host01',
                'hypervisor_type': 'QEMU',
                'hypervisor_version': 1000000,
                'service_name': 'host02',
                'vcpus': 1,
                'memory_mb': 8192,
                'local_gb': 50,
                'cpu_info': "{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                            "\"arch\": \"x86_64\", \"features\": [],"
                            " \"topology\": {\"cores\": 1}}",
                'extra_capas': {'vgpus': 2},
                'trust_id': None}

        if engine.name != 'sqlite':
            # pylint: disable=E1120
            self.assertRaises(sqlalchemy.exc.DBAPIError,
                              engine.execute,
                              computehosts_table.insert(),
                              data)



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Policy Engine For Blazar."""

import functools

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import opts
from oslo_policy import policy

from blazar import context
from blazar import exceptions
from blazar import policies

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


# TODO(gmann): Remove setting the default value of config policy_file
# once oslo_policy change the default value to 'policy.yaml'.
# https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
DEFAULT_POLICY_FILE = 'policy.yaml'
opts.set_defaults(CONF, DEFAULT_POLICY_FILE)


_ENFORCER = None


def reset():
    global _ENFORCER

    if _ENFORCER:
        _ENFORCER.clear()
    _ENFORCER = None


def init():
    global _ENFORCER
    if not _ENFORCER:
        LOG.debug("Enforcer not present, recreating at init stage.")
        _ENFORCER = policy.Enforcer(CONF)
        _ENFORCER.register_defaults(policies.list_rules())


def set_rules(data, default_rule=None):
    default_rule = default_rule or CONF.policy_default_rule
    if not _ENFORCER:
        LOG.debug("Enforcer not present, recreating at rules stage.")
        init()
    if default_rule:
        _ENFORCER.default_rule = default_rule
    _ENFORCER.set_rules(policy.Rules.load(data, default_rule))


def enforce(context, action, target, do_raise=True):
    """Verifies that the action is valid on the target in this context.

       :param context: blazar context
       :param action: string representing the action to be checked
           this should be colon separated for clarity.
           i.e. ``compute:create_instance``,
           ``compute:attach_volume``,
           ``volume:attach_volume``
       :param target: dictionary representing the object of the action
           for object creation this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``
       :param do_raise: if True (the default), raises PolicyNotAuthorized;
           if False, returns False

       :raises blazar.exceptions.PolicyNotAuthorized: if verification fails
           and do_raise is True.

       :return: returns a non-False value (not necessarily "True") if
           authorized, and the exact value False if not authorized and
           do_raise is False.
    """

    init()

    credentials = context.to_policy_values()

    # Add the exceptions arguments if asked to do a raise
    extra = {}
    if do_raise:
        extra.update(exc=exceptions.PolicyNotAuthorized, action=action)

    return _ENFORCER.enforce(action, target, credentials, do_raise=do_raise,
                             **extra)


def authorize(extension, action=None, api='blazar', ctx=None,
              target=None):
    def decorator(func):

        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            cur_ctx = ctx or context.current()
            tgt = target or {'project_id': cur_ctx.project_id,
                             'user_id': cur_ctx.user_id}
            if action is None:
                act = '%s:%s' % (api, extension)
            else:
                act = '%s:%s:%s' % (api, extension, action)
            enforce(cur_ctx, act, tgt)
            return func(self, *args, **kwargs)

        return wrapped
    return decorator



# Copyright (c) 2022 Radosław Piliszek <radoslaw.piliszek@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
from unittest import mock

from blazar.enforcement.exceptions import ExternalServiceFilterException
from blazar.enforcement.filters import external_service_filter
from blazar.tests import TestCase

from oslo_config.cfg import CONF


class FakeResponse204():
    status_code = 204


class FakeResponse403Empty():
    status_code = 403
    content = "irrelevant_but_logged"

    @staticmethod
    def json():
        return {}


class FakeJSONDecodeError(Exception):
    pass


class FakeResponse403InvalidJSON():
    status_code = 403
    content = "NOT_JSON"

    @staticmethod
    def json():
        raise FakeJSONDecodeError()


class FakeResponse403WithMessage():
    status_code = 403
    content = "irrelevant"

    @staticmethod
    def json():
        return {"message": "Hello!"}


class FakeResponse500():
    status_code = 500
    content = "ALL_YOUR_BUGS_BELONG_TO_US"


class ISODateTimeEncoderTestCase(TestCase):

    def test_json_date(self):
        dt = datetime.datetime(2022, 9, 8, 13, 31, 44, 12345)
        obj = {"datetime": dt}
        x = json.dumps(obj, cls=external_service_filter.ISODateTimeEncoder)
        self.assertEqual('{"datetime": "2022-09-08T13:31:44.012345"}', x)

    def test_json_with_tz(self):
        tz = datetime.timezone(datetime.timedelta(hours=2))
        dt = datetime.datetime(2022, 9, 8, 13, 31, 44, 12345, tz)
        obj = {"datetime": dt}
        x = json.dumps(obj, cls=external_service_filter.ISODateTimeEncoder)
        self.assertEqual('{"datetime": "2022-09-08T13:31:44.012345+02:00"}', x)


class ConfiguringExternalServiceFilterTestCase(TestCase):
    def setUp(self):
        super().setUp()

        external_service_filter.ExternalServiceFilter.register_opts(CONF)

    def test_basic_misconfiguration(self):
        self.assertRaises(external_service_filter.ExternalServiceMisconfigured,
                          external_service_filter.ExternalServiceFilter, CONF)

    def test_bad_url(self):
        CONF.set_override(
            'external_service_base_endpoint', 'this_url_cOuLDnOtBeWoRsE',
            group='enforcement')
        self.addCleanup(CONF.clear_override, 'external_service_base_endpoint',
                        group='enforcement')

        self.assertRaises(external_service_filter.ExternalServiceMisconfigured,
                          external_service_filter.ExternalServiceFilter, CONF)

    def test_check_create_endpoint_is_enough(self):
        CONF.set_override(
            'external_service_check_create_endpoint', 'http://localhost',
            group='enforcement')
        self.addCleanup(CONF.clear_override,
                        'external_service_check_create_endpoint',
                        group='enforcement')

        external_service_filter.ExternalServiceFilter(CONF)

    def test_check_updaye_endpoint_is_enough(self):
        CONF.set_override(
            'external_service_check_update_endpoint', 'http://localhost',
            group='enforcement')
        self.addCleanup(CONF.clear_override,
                        'external_service_check_update_endpoint',
                        group='enforcement')

        external_service_filter.ExternalServiceFilter(CONF)

    def test_on_end_endpoint_is_enough(self):
        CONF.set_override(
            'external_service_on_end_endpoint', 'http://localhost',
            group='enforcement')
        self.addCleanup(CONF.clear_override,
                        'external_service_on_end_endpoint',
                        group='enforcement')

        external_service_filter.ExternalServiceFilter(CONF)


class ExternalServiceFilterTestCase(TestCase):
    def setUp(self):
        super().setUp()

        external_service_filter.ExternalServiceFilter.register_opts(CONF)

        CONF.set_override(
            'external_service_base_endpoint', 'http://localhost',
            group='enforcement')
        self.addCleanup(CONF.clear_override, 'external_service_base_endpoint',
                        group='enforcement')

        self.filter = external_service_filter.ExternalServiceFilter(CONF)

        self.ctx = {
            "is_context": True
        }

        self.lease = {
            "is_lease": True
        }

        self.old_lease = {
            "is_old_lease": True
        }

    @mock.patch("requests.post")
    def test_check_create_allowed(self, post_mock):
        post_mock.return_value = FakeResponse204()
        self.filter.check_create(self.ctx, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-create",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_check_create_denied(self, post_mock):
        post_mock.return_value = FakeResponse403WithMessage()
        self.assertRaises(ExternalServiceFilterException,
                          self.filter.check_create,
                          self.ctx, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-create",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_check_create_failed(self, post_mock):
        post_mock.return_value = FakeResponse403Empty()
        self.assertRaises(ExternalServiceFilterException,
                          self.filter.check_create,
                          self.ctx, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-create",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_check_update_allowed(self, post_mock):
        post_mock.return_value = FakeResponse204()
        self.filter.check_update(self.ctx, self.old_lease, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-update",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"current_lease": {"is_old_lease": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_check_update_denied(self, post_mock):
        post_mock.return_value = FakeResponse403WithMessage()
        self.assertRaises(ExternalServiceFilterException,
                          self.filter.check_update,
                          self.ctx, self.old_lease, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-update",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"current_lease": {"is_old_lease": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    @mock.patch("requests.JSONDecodeError", FakeJSONDecodeError)
    def test_check_update_failed(self, post_mock):
        post_mock.return_value = FakeResponse403InvalidJSON()
        self.assertRaises(ExternalServiceFilterException,
                          self.filter.check_update,
                          self.ctx, self.old_lease, self.lease)
        post_mock.assert_called_with(
            "http://localhost/check-update",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"current_lease": {"is_old_lease": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_on_end_success(self, post_mock):
        post_mock.return_value = FakeResponse204()
        self.filter.on_end(self.ctx, self.lease)
        post_mock.assert_called_with(
            "http://localhost/on-end",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"lease": {"is_lease": true}}')

    @mock.patch("requests.post")
    def test_on_end_failure(self, post_mock):
        post_mock.return_value = FakeResponse500()
        self.assertRaises(ExternalServiceFilterException,
                          self.filter.on_end,
                          self.ctx, self.lease)
        post_mock.assert_called_with(
            "http://localhost/on-end",
            headers={'Content-Type': 'application/json'},
            data='{"context": {"is_context": true}, '
                 '"lease": {"is_lease": true}}')



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystonemiddleware import auth_token
from oslo_config import cfg
from oslo_middleware import debug
import pecan

from blazar.api.v2 import hooks
from blazar.api.v2 import middleware


auth_opts = [
    cfg.StrOpt('auth_strategy',
               default='keystone',
               help='The strategy to use for auth: noauth or keystone.'),
]

CONF = cfg.CONF
CONF.register_opts(auth_opts)

CONF.import_opt('log_exchange', 'blazar.config')

OPT_GROUP_NAME = 'keystone_authtoken'


def setup_app(pecan_config=None, extra_hooks=None):

    app_hooks = [hooks.ConfigHook(),
                 hooks.DBHook(),
                 hooks.ContextHook(),
                 hooks.RPCHook(),
                 ]
    # TODO(sbauza): Add stevedore extensions for loading hooks
    if extra_hooks:
        app_hooks.extend(extra_hooks)

    app = pecan.make_app(
        pecan_config.app.root,
        debug=CONF.debug,
        hooks=app_hooks,
        wrap_app=middleware.ParsableErrorMiddleware,
        guess_content_type_from_ext=False
    )

    # WSGI middleware for Keystone auth
    # NOTE(sbauza): ACLs are always active unless for unittesting where
    #               enable_acl could be set to False
    if pecan_config.app.enable_acl:
        app = auth_token.AuthProtocol(app, {})

    return app


def make_app():
    config = {
        'app': {
            'modules': ['blazar.api.v2'],
            'root': 'blazar.api.root.RootController',
            'enable_acl': True,
        }
    }
    # NOTE(sbauza): Fill Pecan config and call modules' path app.setup_app()
    app = pecan.load_app(config)
    # WSGI middleware for debugging
    if CONF.log_exchange:
        app = debug.Debug.factory(config)(app)
    return app



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add_az_in_compute

Revision ID: 35b314cd39ee
Revises: c0ae6b08b0d7
Create Date: 2018-04-23 07:40:39.750686

"""

# revision identifiers, used by Alembic.
revision = '35b314cd39ee'
down_revision = 'c0ae6b08b0d7'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('computehosts',
                  sa.Column('availability_zone', sa.String(length=255),
                            default="", nullable=False))


def downgrade():
    op.drop_column('computehosts', 'availability_zone')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import datetime
from operator import itemgetter

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception
from stevedore import enabled

from blazar import context
from blazar.db import api as db_api
from blazar.db import exceptions as db_ex
from blazar import enforcement
from blazar import exceptions as common_ex
from blazar import manager
from blazar.manager import exceptions
from blazar import monitor
from blazar.notification import api as notification_api
from blazar import status
from blazar.utils import service as service_utils
from blazar.utils import trusts

manager_opts = [
    cfg.ListOpt('plugins',
                default=['dummy.vm.plugin'],
                help='All plugins to use (one for every resource type to '
                     'support.)'),
    cfg.IntOpt('minutes_before_end_lease',
               default=60,
               min=0,
               help='Minutes prior to the end of a lease in which actions '
                    'like notification and snapshot are taken. If this is '
                    'set to 0, then these actions are not taken.'),
    cfg.IntOpt('event_max_retries',
               default=1,
               min=0,
               max=50,
               help='Number of times to retry an event action.')
]

CONF = cfg.CONF
CONF.register_opts(manager_opts, 'manager')
LOG = logging.getLogger(__name__)

LEASE_DATE_FORMAT = "%Y-%m-%d %H:%M"

EVENT_INTERVAL = 10


class ManagerService(service_utils.RPCServer):
    """Service class for the blazar-manager service.

    Responsible for working with Blazar DB, scheduling logic, running events,
    working with plugins, etc.
    """

    def __init__(self):
        target = manager.get_service_target()
        super(ManagerService, self).__init__(target)
        self.plugins = self._get_plugins()
        self.resource_actions = self._setup_actions()
        self.monitors = monitor.load_monitors(self.plugins)
        self.enforcement = enforcement.UsageEnforcement()

    def start(self):
        super(ManagerService, self).start()
        # NOTE(jakecoll): stop_on_exception=False was added because database
        # exceptions would prevent threads from being scheduled again.
        # TODO(jakecoll): Find a way to test this.
        self.tg.add_timer_args(EVENT_INTERVAL, self._process_events,
                               stop_on_exception=False)
        for m in self.monitors:
            m.start_monitoring()

    def _get_plugins(self):
        """Return dict of resource-plugin class pairs."""
        config_plugins = CONF.manager.plugins
        plugins = {}

        extension_manager = enabled.EnabledExtensionManager(
            check_func=lambda ext: ext.name in config_plugins,
            namespace='blazar.resource.plugins',
            invoke_on_load=False
        )

        invalid_plugins = (set(config_plugins) -
                           set([ext.name for ext
                                in extension_manager.extensions]))
        if invalid_plugins:
            raise common_ex.BlazarException('Invalid plugin names are '
                                            'specified: %s' % invalid_plugins)

        for ext in extension_manager.extensions:
            try:
                plugin_obj = ext.plugin()
            except Exception as e:
                LOG.warning("Could not load %s plugin "
                            "for resource type %s '%s'",
                            ext.name, ext.plugin.resource_type, e)
            else:
                if plugin_obj.resource_type in plugins:
                    msg = ("You have provided several plugins for "
                           "one resource type in configuration file. "
                           "Please set one plugin per resource type.")
                    raise exceptions.PluginConfigurationError(error=msg)

                plugins[plugin_obj.resource_type] = plugin_obj
        return plugins

    def _setup_actions(self):
        """Setup actions for each resource type supported.

        BasePlugin interface provides only on_start and on_end behaviour now.
        If there are some configs needed by plugin, they should be returned
        from get_plugin_opts method. These flags are registered in
        [resource_type] group of configuration file.
        """
        actions = {}

        for resource_type, plugin in self.plugins.items():
            plugin = self.plugins[resource_type]
            CONF.register_opts(plugin.get_plugin_opts(), group=resource_type)

            actions[resource_type] = {}
            actions[resource_type]['on_start'] = plugin.on_start
            actions[resource_type]['on_end'] = plugin.on_end
            actions[resource_type]['before_end'] = plugin.before_end
            plugin.setup(None)
        return actions

    @service_utils.with_empty_context
    def _process_events_concurrently(self, events):
        if not events:
            return

        LOG.info("Trying to execute events: %s", events)
        event_threads = {}
        for event in events:
            if not status.LeaseStatus.is_stable(event['lease_id']):
                LOG.info("Skip event %s because the status of the lease %s "
                         "is still transitional", event, event['lease_id'])
                continue
            db_api.event_update(event['id'],
                                {'status': status.event.IN_PROGRESS})
            try:
                event_thread = eventlet.spawn(
                    service_utils.with_empty_context(self._exec_event),
                    event)
                event_threads[event['id']] = event_thread
            except Exception:
                db_api.event_update(event['id'],
                                    {'status': status.event.ERROR})
                LOG.exception('Error occurred while spawning event %s.',
                              event['id'])

        for event_id, event_thread in event_threads.items():
            try:
                event_thread.wait()
            except Exception:
                db_api.event_update(event['id'],
                                    {'status': status.event.ERROR})
                LOG.exception('Error occurred while handling event %s.',
                              event_id)

    def _select_for_execution(self, events):
        """Orders the events such that they can be safely executed concurrently

        Events are selected to be executed concurrently if they are of the same
        type, while keeping strict time ordering and the following priority of
        event types: before_end_lease, end_lease, and start_lease (except for
        before_end_lease events where there is a start_lease event for the same
        lease at the same time).

        We ensure that:

        - the before_end_lease event of a lease is executed after the
          start_lease event and before the end_lease event of the same lease,
        - for two reservations using the same hosts back to back, the end_lease
          event is executed before the start_lease event.
        """
        if not events:
            return []

        events_by_lease = defaultdict(list)
        events_by_type = defaultdict(list)

        for e in sorted(events, key=itemgetter('time')):
            events_by_lease[e['lease_id']].append(e)
            events_by_type[e['event_type']].append(e)

        # If there is a start_lease event for the same lease, we run it first.
        deferred_before_end_events = []
        deferred_end_events = []
        for start_event in events_by_type['start_lease']:
            for e in events_by_lease[start_event['lease_id']]:
                if e['event_type'] == 'before_end_lease':
                    events_by_type['before_end_lease'].remove(e)
                    deferred_before_end_events.append(e)
                elif e['event_type'] == 'end_lease':
                    events_by_type['end_lease'].remove(e)
                    deferred_end_events.append(e)

        return [
            events_by_type['before_end_lease'],
            events_by_type['end_lease'],
            events_by_type['start_lease'],
            deferred_before_end_events,
            deferred_end_events
        ]

    def _process_events(self):
        """Tries to execute events.

        If there is any event in Blazar DB to be executed, do it and change its
        status to 'DONE'. Events are executed concurrently if possible.
        """
        LOG.debug('Trying to get events from DB.')
        events = db_api.event_get_all_sorted_by_filters(
            sort_key='time',
            sort_dir='asc',
            filters={'status': status.event.UNDONE,
                     'time': {'op': 'le',
                              'border': datetime.datetime.utcnow()}}
        )

        for batch in self._select_for_execution(events):
            self._process_events_concurrently(batch)

    def _exec_event(self, event):
        """Execute an event function"""
        event_fn = getattr(self, event['event_type'], None)
        if event_fn is None:
            raise exceptions.EventError(
                error='Event type %s is not supported'
                      % event['event_type'])
        try:
            event_fn(lease_id=event['lease_id'], event_id=event['id'])
        except common_ex.InvalidStatus:
            now = datetime.datetime.utcnow()
            if now < event['time'] + datetime.timedelta(
                    seconds=CONF.manager.event_max_retries * 10):
                # Set the event status UNDONE for retrying the event
                db_api.event_update(event['id'],
                                    {'status': status.event.UNDONE})
            else:
                db_api.event_update(event['id'],
                                    {'status': status.event.ERROR})
                LOG.exception('Error occurred while handling %s event for '
                              'lease %s.', event['event_type'],
                              event['lease_id'])
        except Exception:
            db_api.event_update(event['id'],
                                {'status': status.event.ERROR})
            LOG.exception('Error occurred while handling %s event for '
                          'lease %s.', event['event_type'], event['lease_id'])
        else:
            lease = db_api.lease_get(event['lease_id'])
            with trusts.create_ctx_from_trust(lease['trust_id']) as ctx:
                self._send_notification(
                    lease, ctx, events=['event.%s' % event['event_type']])

    def _date_from_string(self, date_string, date_format=LEASE_DATE_FORMAT):
        try:
            date = datetime.datetime.strptime(date_string, date_format)
        except ValueError:
            raise exceptions.InvalidDate(date=date_string,
                                         date_format=date_format)
        return date

    def _parse_lease_dates(self, start_date, end_date):
        now = datetime.datetime.utcnow()
        now = datetime.datetime(now.year,
                                now.month,
                                now.day,
                                now.hour,
                                now.minute)
        if start_date == 'now':
            start_date = now
        else:
            start_date = self._date_from_string(start_date)

        if end_date == 'now':
            end_date = now
        else:
            end_date = self._date_from_string(end_date)

        return start_date, end_date, now

    def _check_for_invalid_date_inputs(self, lease, values, now):
        if (lease['start_date'] < now and
                values['start_date'] != lease['start_date']):
            raise common_ex.InvalidInput(
                'Cannot modify the start date of already started leases')

        if (lease['start_date'] > now and
                values['start_date'] < now):
            raise common_ex.InvalidInput(
                'Start date must be later than current date')

        if lease['end_date'] < now:
            raise common_ex.InvalidInput(
                'Terminated leases can only be renamed')

        if (values['end_date'] < now or
           values['end_date'] < values['start_date']):
            raise common_ex.InvalidInput(
                'End date must be later than current and start date')

    def validate_params(self, values, required_params):
        if isinstance(required_params, list):
            required_params = set(required_params)
        missing_attr = required_params - set(values.keys())
        if missing_attr:
            raise exceptions.MissingParameter(param=', '.join(missing_attr))

    def get_lease(self, lease_id):
        return db_api.lease_get(lease_id)

    def list_leases(self, project_id=None, query=None):
        return db_api.lease_list(project_id)

    def create_lease(self, lease_values):
        """Create a lease with reservations.

        Return either the model of created lease or None if any error.
        """
        lease_values['status'] = status.lease.CREATING

        try:
            trust_id = lease_values.pop('trust_id')
        except KeyError:
            raise exceptions.MissingTrustId()

        self.validate_params(lease_values, ['name', 'start_date', 'end_date'])

        # Remove and keep event and reservation values
        events = lease_values.pop("events", [])
        reservations = lease_values.pop("reservations", [])
        for res in reservations:
            self.validate_params(res, ['resource_type'])

        # Create the lease without the reservations
        start_date, end_date, now = self._parse_lease_dates(
            lease_values['start_date'], lease_values['end_date'])

        if start_date < now:
            raise common_ex.InvalidInput(
                'Start date must be later than current date')

        if end_date <= start_date:
            raise common_ex.InvalidInput(
                'End date must be later than start date.')

        with trusts.create_ctx_from_trust(trust_id) as ctx:
            # NOTE(priteau): We should not get user_id from ctx, because we are
            # in the context of the trustee (blazar user).
            # lease_values['user_id'] is set in blazar/api/v1/service.py
            lease_values['project_id'] = ctx.project_id
            lease_values['start_date'] = start_date
            lease_values['end_date'] = end_date

            allocations = self._allocation_candidates(
                lease_values, reservations)
            try:
                self.enforcement.check_create(
                    context.current(), lease_values, reservations, allocations)
            except common_ex.NotAuthorized as e:
                LOG.error("Enforcement checks failed. %s", str(e))
                raise common_ex.NotAuthorized(e)

            events.append({'event_type': 'start_lease',
                           'time': start_date,
                           'status': status.event.UNDONE})
            events.append({'event_type': 'end_lease',
                           'time': end_date,
                           'status': status.event.UNDONE})

            before_end_date = lease_values.get('before_end_date', None)
            if before_end_date:
                # incoming param. Validation check
                try:
                    before_end_date = self._date_from_string(
                        before_end_date)
                    self._check_date_within_lease_limits(before_end_date,
                                                         lease_values)
                except common_ex.BlazarException as e:
                    LOG.error("Invalid before_end_date param. %s", str(e))
                    raise e
            elif CONF.manager.minutes_before_end_lease > 0:
                delta = datetime.timedelta(
                    minutes=CONF.manager.minutes_before_end_lease)
                before_end_date = lease_values['end_date'] - delta

            if before_end_date:
                event = {'event_type': 'before_end_lease',
                         'status': status.event.UNDONE}
                events.append(event)
                self._update_before_end_event_date(event, before_end_date,
                                                   lease_values)

            try:
                if trust_id:
                    lease_values.update({'trust_id': trust_id})
                lease = db_api.lease_create(lease_values)
                lease_id = lease['id']
            except db_ex.BlazarDBDuplicateEntry:
                LOG.exception('Cannot create a lease - duplicated lease name')
                raise exceptions.LeaseNameAlreadyExists(
                    name=lease_values['name'])
            except db_ex.BlazarDBException:
                with save_and_reraise_exception():
                    LOG.exception('Cannot create a lease')
            else:
                try:
                    for reservation in reservations:
                        reservation['lease_id'] = lease['id']
                        reservation['start_date'] = lease['start_date']
                        reservation['end_date'] = lease['end_date']
                        self._create_reservation(reservation)
                except Exception:
                    with save_and_reraise_exception():
                        LOG.exception("Failed to create reservation for a "
                                      "lease. Rollback the lease and "
                                      "associated reservations")
                        db_api.lease_destroy(lease_id)

                try:
                    for event in events:
                        event['lease_id'] = lease['id']
                        db_api.event_create(event)
                except (exceptions.UnsupportedResourceType,
                        common_ex.BlazarException):
                    with save_and_reraise_exception():
                        LOG.exception("Failed to create event for a lease. "
                                      "Rollback the lease and associated "
                                      "reservations")
                        db_api.lease_destroy(lease_id)

                else:
                    db_api.lease_update(
                        lease_id,
                        {'status': status.lease.PENDING})
                    lease = db_api.lease_get(lease_id)
                    self._send_notification(lease, ctx, events=['create'])
                    return lease

    def _add_resource_type(self, reservations, existing_reservations):
        rsvns_by_id = {}

        for r in existing_reservations:
            rsvns_by_id[r['id']] = r
        for r in reservations:
            if 'resource_type' not in r:
                r['resource_type'] = rsvns_by_id[r['id']]['resource_type']

        return reservations

    @status.lease.lease_status(
        transition=status.lease.UPDATING,
        result_in=status.lease.STABLE,
        non_fatal_exceptions=[
            common_ex.InvalidInput,
            exceptions.InvalidRange,
            exceptions.MissingParameter,
            exceptions.MalformedRequirements,
            exceptions.MalformedParameter,
            exceptions.NotEnoughHostsAvailable,
            exceptions.InvalidDate,
            exceptions.CantUpdateParameter,
            exceptions.InvalidPeriod,
            enforcement.exceptions.MaxLeaseDurationException,
            enforcement.exceptions.ExternalServiceFilterException,
        ]
    )
    def update_lease(self, lease_id, values):
        if not values:
            return db_api.lease_get(lease_id)

        if len(values) == 1 and 'name' in values:
            db_api.lease_update(lease_id, values)
            return db_api.lease_get(lease_id)

        lease = db_api.lease_get(lease_id)
        start_date = values.get(
            'start_date',
            datetime.datetime.strftime(lease['start_date'], LEASE_DATE_FORMAT))
        end_date = values.get(
            'end_date',
            datetime.datetime.strftime(lease['end_date'], LEASE_DATE_FORMAT))
        before_end_date = values.get('before_end_date', None)

        start_date, end_date, now = self._parse_lease_dates(start_date,
                                                            end_date)

        values['start_date'] = start_date
        values['end_date'] = end_date

        self._check_for_invalid_date_inputs(lease, values, now)

        with trusts.create_ctx_from_trust(lease['trust_id']):
            if before_end_date:
                try:
                    before_end_date = self._date_from_string(before_end_date)
                    self._check_date_within_lease_limits(before_end_date,
                                                         values)
                except common_ex.BlazarException as e:
                    LOG.error("Invalid before_end_date param. %s", str(e))
                    raise e

            reservations = values.get('reservations', [])
            existing_reservations = (
                db_api.reservation_get_all_by_lease_id(lease_id))

            try:
                invalid_ids = set([r['id'] for r in reservations]).difference(
                    [r['id'] for r in existing_reservations])
            except KeyError:
                raise exceptions.MissingParameter(param='reservation ID')

            if invalid_ids:
                raise common_ex.InvalidInput(
                    'Please enter valid reservation IDs. Invalid reservation '
                    'IDs are: %s' % ','.join([str(id) for id in invalid_ids]))

            # Populate resource_type if missing from data passed by client
            reservations = self._add_resource_type(reservations,
                                                   existing_reservations)

            try:
                [
                    self.plugins[r['resource_type']] for r
                    in (reservations + existing_reservations)]
            except KeyError:
                raise exceptions.CantUpdateParameter(param='resource_type')

            existing_allocs = self._existing_allocations(existing_reservations)

            if reservations:
                new_reservations = reservations
                new_allocs = self._allocation_candidates(values,
                                                         existing_reservations)
            else:
                # User is not updating reservation parameters, e.g., is only
                # adjusting lease start/end dates.
                new_reservations = existing_reservations
                new_allocs = existing_allocs

            try:
                self.enforcement.check_update(context.current(), lease, values,
                                              existing_allocs, new_allocs,
                                              existing_reservations,
                                              new_reservations)
            except common_ex.NotAuthorized as e:
                LOG.error("Enforcement checks failed. %s", str(e))
                raise e

            # TODO(frossigneux) rollback if an exception is raised
            for reservation in existing_reservations:
                v = {}
                v['start_date'] = values['start_date']
                v['end_date'] = values['end_date']
                try:
                    v.update([r for r in reservations
                              if r['id'] == reservation['id']].pop())
                except IndexError:
                    pass
                resource_type = v.get('resource_type',
                                      reservation['resource_type'])

                if resource_type != reservation['resource_type']:
                    raise exceptions.CantUpdateParameter(
                        param='resource_type')
                self.plugins[resource_type].update_reservation(
                    reservation['id'], v)

        event = db_api.event_get_first_sorted_by_filters(
            'lease_id',
            'asc',
            {
                'lease_id': lease_id,
                'event_type': 'start_lease'
            }
        )
        if not event:
            raise common_ex.BlazarException(
                'Start lease event not found')
        db_api.event_update(event['id'], {'time': values['start_date']})

        event = db_api.event_get_first_sorted_by_filters(
            'lease_id',
            'asc',
            {
                'lease_id': lease_id,
                'event_type': 'end_lease'
            }
        )
        if not event:
            raise common_ex.BlazarException(
                'End lease event not found')
        db_api.event_update(event['id'], {'time': values['end_date']})

        notifications = ['update']
        self._update_before_end_event(lease, values, notifications,
                                      before_end_date)

        try:
            del values['reservations']
        except KeyError:
            pass
        db_api.lease_update(lease_id, values)

        lease = db_api.lease_get(lease_id)
        with trusts.create_ctx_from_trust(lease['trust_id']) as ctx:
            self._send_notification(lease, ctx, events=notifications)

        return lease

    @status.lease.lease_status(transition=status.lease.DELETING,
                               result_in=(status.lease.ERROR,))
    def delete_lease(self, lease_id):
        lease = self.get_lease(lease_id)

        start_event = db_api.event_get_first_sorted_by_filters(
            'lease_id',
            'asc',
            {
                'lease_id': lease_id,
                'event_type': 'start_lease',
            }
        )
        if not start_event:
            raise common_ex.BlazarException(
                'start_lease event for lease %s not found' % lease_id)

        end_event = db_api.event_get_first_sorted_by_filters(
            'lease_id',
            'asc',
            {
                'lease_id': lease_id,
                'event_type': 'end_lease',
            }
        )
        if not end_event:
            raise common_ex.BlazarException(
                'end_lease event for lease %s not found' % lease_id)

        lease_already_started = start_event['status'] != status.event.UNDONE
        lease_not_started = not lease_already_started
        lease_already_ended = end_event['status'] != status.event.UNDONE
        lease_not_ended = not lease_already_ended

        end_lease = lease_already_started and lease_not_ended

        if end_lease:
            db_api.event_update(end_event['id'],
                                {'status': status.event.IN_PROGRESS})

        with trusts.create_ctx_from_trust(lease['trust_id']) as ctx:
            reservations = lease['reservations']

            if lease_not_started or lease_not_ended:
                # Only run the on_end enforcement if we're explicitly
                # ending the lease for the first time OR if we're terminating
                # it before the lease ever started. It's important to run
                # on_end in the second case to inform enforcement that the
                # lease is no longer in play.
                allocations = self._existing_allocations(reservations)
                try:
                    self.enforcement.on_end(ctx, lease, allocations)
                except Exception as e:
                    LOG.error(e)

            for reservation in reservations:
                if reservation['status'] != status.reservation.DELETED:
                    plugin = self.plugins[reservation['resource_type']]
                    try:
                        plugin.on_end(reservation['resource_id'])
                    except (db_ex.BlazarDBException, RuntimeError):
                        with save_and_reraise_exception():
                            LOG.exception("Failed to delete a reservation "
                                          "for a lease.")
            db_api.lease_destroy(lease_id)
            self._send_notification(lease, ctx, events=['delete'])

    @status.lease.lease_status(
        transition=status.lease.STARTING,
        result_in=(status.lease.ACTIVE, status.lease.ERROR))
    def start_lease(self, lease_id, event_id):
        lease = self.get_lease(lease_id)
        with trusts.create_ctx_from_trust(lease['trust_id']):
            self._basic_action(lease_id, event_id, 'on_start',
                               status.reservation.ACTIVE)

    @status.lease.lease_status(
        transition=status.lease.TERMINATING,
        result_in=(status.lease.TERMINATED, status.lease.ERROR))
    def end_lease(self, lease_id, event_id):
        lease = self.get_lease(lease_id)

        with trusts.create_ctx_from_trust(lease['trust_id']) as ctx:
            allocations = self._existing_allocations(lease['reservations'])
            try:
                self.enforcement.on_end(ctx, lease, allocations)
            except Exception as e:
                LOG.error(e)

            self._basic_action(lease_id, event_id, 'on_end',
                               status.reservation.DELETED)

    def before_end_lease(self, lease_id, event_id):
        lease = self.get_lease(lease_id)
        with trusts.create_ctx_from_trust(lease['trust_id']):
            self._basic_action(lease_id, event_id, 'before_end')

    def _basic_action(self, lease_id, event_id, action_time,
                      reservation_status=None):
        """Commits basic lease actions such as starting and ending."""
        lease = self.get_lease(lease_id)

        event_status = status.event.DONE
        for reservation in lease['reservations']:
            resource_type = reservation['resource_type']
            try:
                if reservation_status is not None:
                    if not status.reservation.is_valid_transition(
                            reservation['status'], reservation_status):
                        raise common_ex.InvalidStatus
                self.resource_actions[resource_type][action_time](
                    reservation['resource_id']
                )
            except common_ex.BlazarException:
                LOG.exception("Failed to execute action %(action)s "
                              "for lease %(lease)s",
                              {'action': action_time,
                               'lease': lease_id})
                event_status = status.event.ERROR
                db_api.reservation_update(
                    reservation['id'],
                    {'status': status.reservation.ERROR})
            else:
                if reservation_status is not None:
                    db_api.reservation_update(reservation['id'],
                                              {'status': reservation_status})

        db_api.event_update(event_id, {'status': event_status})

        return event_status

    def _create_reservation(self, values):
        resource_type = values['resource_type']
        if resource_type not in self.plugins:
            raise exceptions.UnsupportedResourceType(
                resource_type=resource_type)
        reservation_values = {
            'lease_id': values['lease_id'],
            'resource_type': resource_type,
            'status': status.reservation.PENDING
        }
        reservation = db_api.reservation_create(reservation_values)
        resource_id = self.plugins[resource_type].reserve_resource(
            reservation['id'],
            values
        )
        db_api.reservation_update(reservation['id'],
                                  {'resource_id': resource_id})

    def _allocation_candidates(self, lease, reservations):
        """Returns dict by resource type of reservation candidates."""
        allocations = {}

        for reservation in reservations:
            res = reservation.copy()
            resource_type = reservation['resource_type']
            res['start_date'] = lease['start_date']
            res['end_date'] = lease['end_date']

            if resource_type not in self.plugins:
                raise exceptions.UnsupportedResourceType(
                    resource_type=resource_type)

            plugin = self.plugins.get(resource_type)

            if not plugin:
                raise common_ex.BlazarException(
                    'Invalid plugin names are specified: %s' % resource_type)

            candidate_ids = plugin.allocation_candidates(res)

            allocations[resource_type] = [
                plugin.get(cid) for cid in candidate_ids]

        return allocations

    def _existing_allocations(self, reservations):
        allocations = {}

        for reservation in reservations:
            resource_type = reservation['resource_type']

            if resource_type not in self.plugins:
                raise exceptions.UnsupportedResourceType(
                    resource_type=resource_type)

            plugin = self.plugins.get(resource_type)

            if not plugin:
                raise common_ex.BlazarException(
                    'Invalid plugin names are specified: %s' % resource_type)

            resource_ids = [
                x['resource_id'] for x in plugin.list_allocations(
                    dict(reservation_id=reservation['id']))
                if x['reservations']]

            allocations[resource_type] = [
                plugin.get(rid) for rid in resource_ids]

        return allocations

    def _send_notification(self, lease, ctx, events=[]):
        payload = notification_api.format_lease_payload(lease)

        for event in events:
            notification_api.send_lease_notification(ctx, payload,
                                                     'lease.%s' % event)

    def _check_date_within_lease_limits(self, date, lease):
        if not lease['start_date'] < date < lease['end_date']:
            raise common_ex.NotAuthorized(
                'Datetime is out of lease limits')

    def _update_before_end_event_date(self, event, before_end_date, lease):
        event['time'] = before_end_date
        if event['time'] < lease['start_date']:
            LOG.warning("Start_date greater than before_end_date. "
                        "Setting before_end_date to %(start_date)s for "
                        "lease %(id_name)s",
                        {'start_date': lease['start_date'],
                         'id_name': lease.get('id', lease.get('name'))})
            event['time'] = lease['start_date']

    def _update_before_end_event(self, old_lease, new_lease,
                                 notifications, before_end_date=None):
        event = db_api.event_get_first_sorted_by_filters(
            'lease_id',
            'asc',
            {
                'lease_id': old_lease['id'],
                'event_type': 'before_end_lease'
            }
        )
        if event:
            # NOTE(casanch1) do nothing if the event does not exist.
            # This is for backward compatibility
            update_values = {}
            if not before_end_date:
                # before_end_date needs to be calculated based on
                # previous delta
                prev_before_end_delta = old_lease['end_date'] - event['time']
                before_end_date = new_lease['end_date'] - prev_before_end_delta

            self._update_before_end_event_date(update_values, before_end_date,
                                               new_lease)
            if event['status'] == status.event.DONE:
                update_values['status'] = status.event.UNDONE
                notifications.append('event.before_end_lease.stop')

            db_api.event_update(event['id'], update_values)

    def __getattr__(self, name):
        """RPC Dispatcher for plugins methods."""

        fn = None
        try:
            resource_type, method = name.rsplit(':', 1)
        except ValueError:
            # NOTE(sbauza) : the dispatcher needs to know which plugin to use,
            #  raising error if consequently not
            raise AttributeError(name)
        try:
            try:
                fn = getattr(self.plugins[resource_type], method)
            except KeyError:
                LOG.error("Plugin with resource type %s not found",
                          resource_type)
                raise exceptions.UnsupportedResourceType(
                    resource_type=resource_type)
        except AttributeError:
            LOG.error("Plugin %s doesn't include method %s",
                      self.plugins[resource_type], method)
        if fn is not None:
            return fn
        raise AttributeError(name)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
import oslo_messaging as messaging

opts = [
    cfg.StrOpt('rpc_topic',
               default='blazar.manager',
               help='The topic Blazar uses for blazar-manager messages.'),
]

CONF = cfg.CONF
CONF.register_opts(opts, 'manager')
CONF.import_opt('host', 'blazar.config')
RPC_API_VERSION = '1.0'


def get_target():
    return messaging.Target(topic=CONF.manager.rpc_topic,
                            version=RPC_API_VERSION,
                            namespace='manager.api')


def get_service_target():
    return messaging.Target(topic=CONF.manager.rpc_topic,
                            version=RPC_API_VERSION,
                            server=CONF.host,
                            namespace='manager.api')



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar import exceptions
from blazar import tests


class BlazarExceptionTestCase(tests.TestCase):
    def test_default_error_msg(self):
        class FakeBlazarException(exceptions.BlazarException):
            msg_fmt = "default message"

        exc = FakeBlazarException()
        self.assertEqual('default message', str(exc))

    def test_error_msg(self):
        self.assertEqual('test',
                         str(exceptions.BlazarException('test')))

    def test_default_error_msg_with_kwargs(self):
        class FakeBlazarException(exceptions.BlazarException):
            msg_fmt = "default message: %(code)s"

        exc = FakeBlazarException(code=500)
        self.assertEqual('default message: 500', str(exc))

    def test_error_msg_exception_with_kwargs(self):
        class FakeBlazarException(exceptions.BlazarException):
            msg_fmt = "default message: %(mispelled_code)s"

        exc = FakeBlazarException(code=500, mispelled_code='blah')
        self.assertEqual('default message: blah', str(exc))

    def test_default_error_code(self):
        class FakeBlazarException(exceptions.BlazarException):
            code = 404

        exc = FakeBlazarException()
        self.assertEqual(404, exc.kwargs['code'])

    def test_error_code_from_kwarg(self):
        class FakeBlazarException(exceptions.BlazarException):
            code = 500

        exc = FakeBlazarException(code=404)
        self.assertEqual(404, exc.kwargs['code'])

    def test_policynotauthorized_exception(self):
        exc = exceptions.PolicyNotAuthorized(action='foo')
        self.assertEqual("Policy doesn't allow foo to be performed",
                         str(exc))
        self.assertEqual(403, exc.kwargs['code'])



# Copyright 2017 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add fields for resource monitoring

Revision ID: 97d2cad1504e
Revises: 6bfd1c23aa18
Create Date: 2017-11-01 13:25:20.479355

"""

# revision identifiers, used by Alembic.
revision = '97d2cad1504e'
down_revision = '6bfd1c23aa18'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('computehosts', sa.Column('reservable', sa.Boolean(),
                                            server_default=sa.true(),
                                            nullable=False))
    op.add_column('leases', sa.Column('degraded', sa.Boolean(),
                                      server_default=sa.false(),
                                      nullable=False))
    op.add_column('reservations', sa.Column('missing_resources',
                                            sa.Boolean(),
                                            server_default=sa.false(),
                                            nullable=False))
    op.add_column('reservations', sa.Column('resources_changed',
                                            sa.Boolean(),
                                            server_default=sa.false(),
                                            nullable=False))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('reservations', 'resources_changed')
    op.drop_column('reservations', 'missing_resources')
    op.drop_column('leases', 'degraded')
    op.drop_column('computehosts', 'reservable')
    # ### end Alembic commands ###



# -*- encoding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for ACL. Checks whether certain kinds of requests
are blocked or allowed to be processed.
"""
from blazar import policy
from blazar.tests import api


class TestACL(api.APITest):

    def setUp(self):
        super(TestACL, self).setUp()

        self.policy = policy
        self.path = '/os-hosts'
        self.patch(
            self.hosts_rpcapi, 'list_computehosts').return_value = []

    def _make_app(self):
        return super(TestACL, self)._make_app(enable_acl=True)

    def test_non_authenticated(self):
        response = self.get_json(self.path, expect_errors=True)
        self.assertEqual(401, response.status_int)

    def test_authenticated(self):
        response = self.get_json(self.path,
                                 headers={'X-Auth-Token': self.ADMIN_TOKEN})

        self.assertEqual([], response)

    def test_non_admin(self):
        response = self.get_json(self.path,
                                 headers={'X-Auth-Token': self.MEMBER_TOKEN},
                                 expect_errors=True)

        self.assertEqual(403, response.status_int)

    def test_non_admin_with_admin_header(self):
        response = self.get_json(self.path,
                                 headers={'X-Auth-Token': self.MEMBER_TOKEN,
                                          'X-Roles': 'admin'},
                                 expect_errors=True)

        self.assertEqual(403, response.status_int)
